from grins.gen_params import gen_param_names  # noqa: F401


# Internal function to generate a hills equation string of the edge
def _gen_edge_hills(edge):
    """
    Generate the edge hills function based on the edge type.

    Args:
        edge (dict): The edge dictionary (Series) containing the source, target, and type information.

    Returns:
        str: The generated edge hills function.

    """
    sn, tn, ia = edge["Source"], edge["Target"], edge["Type"]
    if ia == 1:
        # print(f"psHill(ActFld_{sn}_{tn}, Thr_{sn}_{tn}, Hill_{sn}_{tn})")
        return f"regfn.psH({sn}, ActFld_{sn}_{tn}, Thr_{sn}_{tn}, Hill_{sn}_{tn})"
    else:
        # print(f"nsHill(InhFld_{sn}_{tn}, Thr_{sn}_{tn}, Hill_{sn}_{tn})")
        return f"regfn.nsH({sn}, InhFld_{sn}_{tn}, Thr_{sn}_{tn}, Hill_{sn}_{tn})"


# Function to take in the target rwos and generate the ODE for a node
def gen_node_ode(target_edges, node_name):
    """
    Generate the ordinary differential equation (ODE) for a given node.

    Parameters:
    - target_edges: DataFrame containing the target edges of the node.
    - node_name: Name of the node.

    Returns:
    - ODE string representing the regulation terms for the node.

    """
    # Check if the target_edges is empty
    if not target_edges.empty:
        # Apply gen_edge_hills to each row and convert the values to a string joined by *
        return f"Prod_{node_name}*{'*'.join(target_edges.apply(_gen_edge_hills, axis=1))} - Deg_{node_name}*{node_name}"
    else:
        # Only Production term - degradation term
        return f"Prod_{node_name} - Deg_{node_name}*{node_name}"


# Function to the generate the ODE file for diffrax from a topo file
def gen_diffrax_odesys(topo_df, topo_name, save_dir="."):
    """
    Generate the ODE system code for diffrax based on the given topology dataframe.

    Args:
        topo_df (pandas.DataFrame): The topology dataframe containing the edges information.
        topo_name (str): The name of the topology.
        save_dir (str, optional): The directory to save the generated code. Defaults to ".".

    Returns:
        None:  Saves the generated file in the driectory specified by save_dir.
    """
    # Get the list of parameters, target nodes and source nodes
    param_names_list, target_nodes, source_nodes = gen_param_names(topo_df)
    # List of unique nodes
    unique_nodes = sorted(set(target_nodes + source_nodes))
    # Inititalise a list to store the ODE strings
    ode_list = [
        "import grins.reg_funcs as regfn\n",
        "def odesys(t,y,args):",
        f"\t({', '.join(unique_nodes)}) = y",
        f"\t({', '.join(param_names_list)}) = args",
    ]
    # Loop through the target nodes
    for ni, nod in enumerate(unique_nodes):
        # Get the edges where n is the target node
        target_edges = topo_df[topo_df["Target"] == nod]
        # The diffrax ODE for each node is d_<nod> = <ODE>
        ode_list.append("\t" + f"d_{nod} = {gen_node_ode(target_edges, nod)}")
    # Append the d_y line
    ode_list.append(f"\td_y = ({', '.join([f'd_{nod}' for nod in unique_nodes])})")
    # Append the end line
    ode_list.append("\treturn d_y\n")
    # Write the lines to a file
    with open(f"{save_dir}/{topo_name}.py", "w") as f:
        f.write("\n".join(ode_list))



import os

# os.environ["XLA_FLAGS"] = "--xla_force_host_platform_device_count=15"
# os.environ["XLA_PYTHON_CLIENT_PREALLOCATE"] = "false"
# os.environ["XLA_PYTHON_CLIENT_MEM_FRACTION"] = ".60"

import jax as jx
import jax.numpy as jnp
import numpy as np

# import matplotlib.pyplot as plt
from diffrax import diffeqsolve, ODETerm, SaveAt, Tsit5, PIDController, SteadyStateEvent  # noqa: E402
from jax import vmap, pmap  # noqa: E402
import pandas as pd
import time

from multiprocessing import Pool

# # Defining a custom addition function
# def add(a, b):
#     return a + b


# DEfining the ODE system
def vector_field(t, y, args):
    prey, predator = y
    α, β, γ, δ = args
    d_prey = α * prey - β * prey * predator
    d_predator = -γ * predator + δ * prey * predator
    # d_predator = add(-γ * predator, δ * prey * predator)
    d_y = d_prey, d_predator
    return d_y


# Defining the shifted hills functions
# Positive Shifted Hill function
def psH(nod, fld, thr, hill):
    return (fld + (1 - fld) * (1 / (1 + jnp.power(nod / thr, hill)))) / fld


# Negative Shifted Hill function
def nsH(nod, fld, thr, hill):
    return fld + (1 - fld) * (1 / (1 + jnp.power(nod / thr, hill)))


# Function for Toggle Switch
def toggle_switch(t, y, args):
    A, B = y
    # (
    #     Prod_A,
    #     Prod_B,
    #     Deg_A,
    #     Deg_B,
    #     InhFld_B_A,
    #     Thr_B_A,
    #     Hill_B_A,
    #     InhFld_A_B,
    #     Thr_A_B,
    #     Hill_A_B,
    # ) = args
    # For RACIPE paramn
    (
        Prod_A,
        Prod_B,
        Deg_A,
        Deg_B,
        Thr_B_A,
        Hill_B_A,
        InhFld_B_A,
        Thr_A_B,
        Hill_A_B,
        InhFld_A_B,
    ) = args
    d_A = Prod_A * nsH(B, InhFld_B_A, Thr_B_A, Hill_B_A) - Deg_A * A
    d_B = Prod_B * nsH(A, InhFld_A_B, Thr_A_B, Hill_A_B) - Deg_B * B
    d_y = d_A, d_B
    return d_y


# Wrapper function to sovle the ODE
def solve_ode(vector_field, solver, t0, t1, dt0, y0, args, saveat, controller):
    # term = ODETerm(vector_field)
    # solver = Tsit5()
    # t0 = 0
    # t1 = 140
    # dt0 = 0.1
    y0 = tuple(y0)
    # args = (0.1, 0.02, 0.4, 0.02)
    # saveat = SaveAt(t1=True)
    sol = diffeqsolve(
        ODETerm(vector_field),
        solver,
        t0,
        t1,
        dt0,
        y0,
        args=args,
        saveat=saveat,
        max_steps=1000,
        stepsize_controller=controller,
        discrete_terminating_event=SteadyStateEvent(),
        # throw=None,
    )
    sol = [s[-1] for s in sol.ys]
    return sol


# term = ODETerm(vector_field)
# solver = Tsit5()
# t0 = 0
# t1 = 140
# dt0 = 0.1
# max_steps = None
# controller = PIDController(rtol=1e-4, atol=1e-6)
# args = (0.1, 0.02, 0.4, 0.02)
# saveat = SaveAt(t1=True)

# # Make a 2xnum_inital of all the random initial conditions
# y0 = jnp.array(np.random.rand(10000, 2) * 10)
# # # Get shape of the initial conditions
# print(y0.shape)

# # Track the time taken to solve the ODEs
# print("Solving the ODEs")
# start = time.time()
# # Map the solve_ode function over the list of initial conditions
# # sol_li = vmap(lambda i: solve_ode(vector_field, t0, t1, dt0, i, args, saveat))(y0)
# sol_li = vmap(
#     lambda i: solve_ode(vector_field, solver, t0, t1, dt0, i, args, saveat, controller),
# )(y0)
# # Print the time taken to solve the ODEs
# print(f"Time taken to solve the ODEs: {time.time() - start}")

# # Convert the solution list to a numpy array
# sol_li = jnp.array(sol_li).T
# # Convert the solution list to a data frame
# sol_df = pd.DataFrame(sol_li, columns=["Prey", "Predator"])
# print(sol_li.shape)
# print(sol_df)
# print(sol_df.describe())


# Defining the problem for the toggle switch
ts_term = ODETerm(toggle_switch)
solver = Tsit5()
t0 = 0
t1 = 140
dt0 = 0.1
max_steps = None
controller = PIDController(rtol=1e-5, atol=1e-6)
# RACIPE parameters: 8.980273,55.355989,0.935186,0.767823,34.754825,5.000000,0.016662,14.209517,4.000000,0.023031
# args = (
#     26.11365514714271,
#     71.88358915410936,
#     0.5858103133738041,
#     0.576099329534918,
#     0.04951568856870413,
#     62.75370424835297,
#     4.0,
#     0.010144997482053913,
#     2.8172455595347636,
#     6.0,
# )
args = (
    8.980273,
    55.355989,
    0.935186,
    0.767823,
    34.754825,
    5.000000,
    0.016662,
    14.209517,
    4.000000,
    0.023031,
)
saveat = SaveAt(t1=True)
# Add steady state temrination event
ss_event = SteadyStateEvent()

# y0 = (40.389327354729176, 60.03178594075143)
# Repeat y0 10 times and get a 2x10 array
# y0 = jnp.array([y0] * 10000)
# Rnaomly generate 10000 initial conditions
y0 = jnp.array(np.random.rand(10000000, 2) * 100)
print(y0.shape)


# Track the time taken to solve the ODEs
print("Solving the ODEs")
start = time.time()
# Map the solve_ode function over the list of initial conditions
# sol_li = vmap(lambda i: solve_ode(vector_field, t0, t1, dt0, i, args, saveat))(y0)
sol_li = vmap(
    lambda i: solve_ode(
        toggle_switch, solver, t0, t1, dt0, i, args, saveat, controller
    ),
)(y0)
# Print the time taken to solve the ODEs
print(f"Time taken to solve the ODEs: {time.time() - start}")

# Convert the solution list to a numpy array
sol_li = jnp.array(sol_li).T
# Convert the solution list to a data frame
sol_df = pd.DataFrame(sol_li, columns=["A", "B"])
print(sol_li.shape)
print(sol_df)
print(sol_df.describe())



from glob import glob
import sys
import os
from grins.gen_diffrax_ode import gen_diffrax_odesys
from grins.gen_params import gen_param_df, gen_init_cond, gen_param_range_df, parse_topos
from importlib import import_module
from diffrax import (
    diffeqsolve,
    ODETerm,
    SaveAt,
    Tsit5,
    PIDController,
    Event,
    steady_state_event,
)
import subprocess
from multiprocessing import Pool  # noqa: F401
import pandas as pd  # noqa: F401
import jax.numpy as jnp
from jax import jit, vmap
import numpy as np
import time
from typing import Union
from scipy.signal import find_peaks


# Function to generate the required directory structure
# def gen_sim_dirstruct(topo_file, save_dir=".", num_replicates=3):
def gen_sim_dirstruct(
    topo_file: str, save_dir: str = ".", num_replicates: int = 3
) -> None:
    """
    Generate directory structure for simulation run.

    Parameters
    -----------
    topo_file : str
        Path to the topo file.
    save_dir : str, optional
        Directory to save the generated structure. Defaults to ".".
    num_replicates : int, optional
        Number of replicates to generate. Defaults to 3.

    Returns
    --------
    None
        Directory structure is created with the topo file name and three folders for the replicates.
    """
    # Get the topo file name
    topo_name = topo_file.split("/")[-1].split(".")[0]
    # Check if the folder with the name of topo file exists
    os.makedirs(f"{save_dir}/{topo_name}", exist_ok=True)
    # Move the topo file to the created folder
    subprocess.run(
        [
            "cp",
            topo_file,
            f"{save_dir}/{topo_name}/{topo_file.split('/')[-1]}",
        ]
    )
    # Make the replicate directories
    for rep in range(1, num_replicates + 1):
        os.makedirs(f"{save_dir}/{topo_name}/{rep:03}", exist_ok=True)
    return None


# Functiont to generate all the parameters related files with replicates
def gen_topo_param_files(
    topo_file: str,
    save_dir: str = ".",
    num_replicates: int = 3,
    num_params: int = 2**10,
    num_init_conds: int = 2**7,
    sampling_method: Union[str, dict] = "Sobol",
):
    """
    Generate parameter files for simulation.

    Parameters
    ----------
    topo_file : str
        The path to the topo file.
    save_dir : str, optional
        The directory where the parameter files will be saved. Defaults to ".".
    num_params : int, optional
        The number of parameter files to generate. Defaults to 2**10.
    num_init_conds : int, optional
        The number of initial condition files to generate. Defaults to 2**7.
    sampling_method : Union[str, dict], optional
        The method to use for sampling the parameter space. Defaults to 'Sobol'. For a finer control over the parameter generation look at the documentation of the gen_param_range_df function and gen_param_df function.

    Returns
    -------
    None
        The parameter files and initial conditions are generated and saved in the specified replicate directories.
    """
    # Get the name of the topo file
    topo_name = topo_file.split("/")[-1].split(".")[0]
    # Parse the topo file
    topo_df = parse_topos(topo_file)
    # # Generate the parameter names
    # param_names = gen_param_names(topo_df)
    # Get the unique nodes in the topo file
    # unique_nodes = sorted(set(param_names[1] + param_names[2]))
    # Generate the required directory structure
    gen_sim_dirstruct(topo_file, save_dir, num_replicates)
    # Specify directory where all the generated ode system file will be saved
    sim_dir = f"{save_dir}/{topo_file.split('/')[-1].split('.')[0]}"
    # Generate the ODE system for diffrax
    gen_diffrax_odesys(topo_df, topo_name, sim_dir)
    # Generate the parameter dataframe and save in each of the replicate folders
    for rep in range(1, num_replicates + 1):
        # Generate the parameter range dataframe
        param_range_df = gen_param_range_df(
            topo_df, num_params, sampling_method=sampling_method
        )
        # Save the parameter range dataframe
        param_range_df.to_csv(
            f"{sim_dir}/{rep:03}/{topo_name}_param_range_{rep:03}.csv",
            index=False,
            sep="\t",
        )
        # # Generate the parameter dataframe with the default values
        param_df = gen_param_df(param_range_df, num_params)
        # print(param_df)
        param_df.to_parquet(
            f"{sim_dir}/{rep:03}/{topo_name}_params_{rep:03}.parquet", index=False
        )
        # Generate the initial conditions dataframe
        initcond_df = gen_init_cond(topo_df=topo_df, num_init_conds=num_init_conds)
        # print(initcond_df)
        initcond_df.to_parquet(
            f"{sim_dir}/{rep:03}/{topo_name}_init_conds_{rep:03}.parquet",
            index=False,
        )
    print(f"Parameter and Intial Condition files generated for {topo_name}")
    return None


# Function to load the ODE system from a specified topology folder as a module so that the ODETerm can be initialized from the ODE system
def load_odeterm(topo_name, simdir):
    """
    Loads an ODE system from a specified topology module and returns an ODETerm object.

    Parameters
    ----------
    topo_name : str
        The name of the topology module to import.
    simdir : str
        The directory path where the topology module is located.

    Returns
    -------
    ODETerm
        An object representing the ODE system.

    Raises
    ------
    ImportError
        If the specified module cannot be imported.
    AttributeError
        If the module does not contain an attribute named 'odesys'.
    """
    sys.path.append(f"{simdir}")
    mod = import_module(f"{topo_name}")
    term = ODETerm(getattr(mod, "odesys"))
    return term


# Function to generate the combinations of initial conditions and parameters
def _gen_combinations(num_init_conds, num_params):
    """
    Generate combinations of initial conditions and parameters.

    This function generates all possible combinations of initial conditions and parameters
    using efficient operations from the JAX library. The combinations are generated as pairs of indices from the initial conditions and parameters dataframes.

    Parameters
    ----------
    num_init_conds : int
        The number of initial conditions.
    num_params : int
        The number of parameters.

    Returns
    -------
    jnp.ndarray
        A 2D array where each row represents a combination of an initial condition and a parameter. The first column contains indices of initial conditions, and the second column contains indices of parameters.
    """
    # Generate the combinations more efficiently using numpy
    i = jnp.repeat(jnp.arange(num_init_conds), num_params)
    p = jnp.tile(jnp.arange(num_params), num_init_conds)
    icprm_comb = jnp.stack([i, p], axis=1)
    return icprm_comb


# Function to parameterise the ODE system and return the right functions
def parameterise_solveode(
    ode_term,
    solver,
    t0,
    t1,
    dt0,
    saveat,
    stepsize_controller,
    max_steps,
    initial_conditions,
    parameters,
):
    """
    Parameterise the ODE system and return the right functions.

    Parameters
    ----------
    term : ODETerm
        The ODE system to solve.
    solver : object
        The solver to be used for solving the ODE.
    t0 : float
        The initial time.
    t1 : float
        The final time.
    dt0 : float
        The initial time step.
    saveat : SaveAt
        The time steps to save the solution at.
    stepsize_controller : PIDController
        The controller for the step sizes.
    max_steps : int
        The maximum number of steps to take.
    initial_conditions : jnp.ndarray
        The initial conditions.
    parameters : jnp.ndarray
        The parameters.

    Returns
    -------
    function
        The functions to solve the ODE system.
    """
    # Check if number of time steps to save is None
    if saveat.subs.ts is None:
        # Function to solve the ODEs
        def solve_steadystate_ode(pi_row):
            sol = diffeqsolve(
                ode_term,  # ODETerm
                solver,  # Solver
                t0,  # Start time
                t1,  # End time
                dt0,  # Time step
                tuple(initial_conditions[pi_row[0]][:-1]),  # Initial conditions
                tuple(parameters[pi_row[1]][:-1]),  # Parameters
                max_steps=max_steps,
                saveat=saveat,
                stepsize_controller=stepsize_controller,
                event=Event(steady_state_event()),
                throw=False,
            )
            # Concatenate the steady state values, time point and event mask
            sol = jnp.concatenate(
                [
                    jnp.array(sol.ys).T,
                    jnp.array(sol.ts[-1])[None, None],
                    jnp.array(sol.event_mask, dtype=jnp.float32)[None, None],
                ],
                axis=1,
            )
            # Add the initial condition and parameter numbers
            sol = jnp.column_stack(
                [
                    sol,
                    initial_conditions[pi_row[0], -1],
                    parameters[pi_row[1], -1],
                ]
            )
            return sol

        # Return the functions to solve the ODEs and format the steady state solutions
        return solve_steadystate_ode

    # If the number of time steps to save is not None
    else:
        # Function to solve the ODEs
        def solve_timeseries_ode(pi_row):
            sol = diffeqsolve(
                ode_term,  # ODETerm
                solver,  # Solver
                t0,  # Start time
                t1,  # End time
                dt0,  # Time step
                tuple(initial_conditions[pi_row[0]][:-1]),  # Initial conditions
                tuple(parameters[pi_row[1]][:-1]),  # Parameters
                max_steps=max_steps,
                saveat=saveat,
                stepsize_controller=stepsize_controller,
                throw=False,
            )
            # Concatenate the time series values and time points
            sol = jnp.concatenate(
                [jnp.array(sol.ys).T, jnp.array(sol.ts)[:, None]], axis=1
            )
            # Add the initial condition and parameter numbers
            sol = jnp.column_stack(
                [
                    sol,
                    jnp.repeat(initial_conditions[pi_row[0], -1], len(sol)),
                    jnp.repeat(parameters[pi_row[1], -1], len(sol)),
                ]
            )
            return sol

        # Return the functions to solve the ODEs and format the time series solutions
        return solve_timeseries_ode


# Functiont to run the simulation for a given topo file
def topo_simulate(
    topo_file,
    replicate_dir,
    initial_conditions,
    parameters,
    t0=0.0,
    tmax=200.0,
    dt0=0.01,
    tsteps=None,
    rel_tol=1e-5,
    abs_tol=1e-6,
    max_steps=2048,
    batch_size=10000,
    ode_term_dir=None,
):
    """
    Simulates the ODE system defined by the topology file and saves the results in the replicate directory. The ode system is loaded as a diffrax ode term and the initial conditions and parameters are passed as jax arrays. The simulation is run for the specified time range and time steps and the results are saved in parquet format in the replicate directory.

    Parameters
    ----------
    topo_file : str
        Path to the topology file.
    replicate_dir : str
        Directory where the replicate results will be saved.
    initial_conditions : pd.DataFrame
        DataFrame containing the initial conditions.
    parameters : pd.DataFrame
        DataFrame containing the parameters.
    t0 : float, optional
        Initial time for the simulation. Default is 0.0.
    tmax : float, optional
        Maximum time for the simulation. Default is 100.0.
    dt0 : float, optional
        Initial time step size. Default is 0.1.
    tsteps : list, optional
        List of time steps at which to save the results. Default is None.
    rel_tol : float, optional
        Relative tolerance for the ODE solver. Default is 1e-5.
    abs_tol : float, optional
        Absolute tolerance for the ODE solver. Default is 1e-6.
    max_steps : int, optional
        Maximum number of steps for the ODE solver. Default is 2048.
    batch_size : int, optional
        Batch size for processing combinations of initial conditions and parameters. Default is 10000.
    ode_term_dir : str, optional
        Directory where the ODE system file is located. Default is None. If None, the parent directory of the replicate directory is assumed to contain the ODE system file. The ODE system file should be named as the topo file with the .py extension.


    Returns:
    -------
    pd.DataFrame
        DataFrame containing the solutions of the ODE system.
    """
    # Get the name of the topo file
    topo_name = topo_file.split("/")[-1].split(".")[0]
    # Making sure to remove the trailing slash from the replicate directory path if it exists
    replicate_dir = replicate_dir.rstrip("/")
    # Check if the ode term directory is None
    if ode_term_dir is None:
        # Getting the parent directory of the replicate directory to get the ODE system file
        simul_dir = os.path.dirname(replicate_dir)
        print(f"Loading ODE system from: {simul_dir}")
    else:
        # Making sure to remove the trailing slash from the ode term directory path if it exists
        simul_dir = ode_term_dir.rstrip("/")
    # Load the ODE system as a diffrax ode term
    ode_term = load_odeterm(topo_name, simul_dir)
    # Getting the intial conditions dataframe column names
    ic_columns = initial_conditions.columns
    # Converting the initial conditions and parameters to jax arrays
    initial_conditions = jnp.array(initial_conditions.to_numpy())
    parameters = jnp.array(parameters.to_numpy())
    # Get the combinations of initial conditions and parameters
    icprm_comb = _gen_combinations(len(initial_conditions), len(parameters))
    print(f"Number of combinations to simulate: {len(icprm_comb)}")
    # Processing the time steps
    if tsteps is None:
        saveat = SaveAt(t1=True)
        print(
            f"Running steady sate simulations for replicate: {replicate_dir.split('/')[-1]}"
        )
    else:
        # Checking if the time steps are in the correct format
        tsteps = sorted(tsteps)
        # If the final step more than tmax, make tmax equal to the final step
        if tsteps[-1] != tmax:
            tmax = tsteps[-1]
        # Convert the time steps to a jax array
        tsteps = jnp.array(tsteps)
        # Make the saveat object be the time steps
        saveat = SaveAt(ts=tsteps)
        print(
            f"Running time series simulations for replicate: {replicate_dir.split('/')[-1]}"
        )
    # Specifying the PID controller for the step sizes
    stepsize_controller = PIDController(rtol=rel_tol, atol=abs_tol)
    # Get the functions to solve the ODE system
    solveode_fn = parameterise_solveode(
        ode_term,
        Tsit5(),
        t0,
        tmax,
        dt0,
        saveat,
        stepsize_controller,
        max_steps,
        initial_conditions,
        parameters,
    )
    # Jit compile the solveode function
    solveode_fn = jit(solveode_fn)
    # # # Solve for one combination of initial conditions and parameters
    # sol = solveode_fn(icprm_comb[0])
    # print(sol)
    # Create an empty array to store the solutions
    if saveat.subs.ts is None:
        solution_matrix = np.zeros(
            (len(icprm_comb), initial_conditions.shape[1] + 3), dtype=np.float32
        )
    else:
        solution_matrix = np.zeros(
            (len(icprm_comb) * len(saveat.subs.ts), initial_conditions.shape[1] + 2),
            dtype=np.float32,
        )
    # Defining the length of the time steps to properly index the solution matrix
    len_tsteps = len(saveat.subs.ts) if saveat.subs.ts is not None else 1
    # Iterate over the combinations array in batches
    for ip in range(0, len(icprm_comb), batch_size):
        # print(ip)
        # Get the chunk of the combinations array
        icprm_chunk = icprm_comb[ip : ip + batch_size]
        # vmap the solveode function over the chunk of the combinations array
        sols = vmap(solveode_fn)(icprm_chunk)
        # Vertically stack the solutions
        sols = jnp.vstack(sols)
        # Round the solutions to 4 decimal places
        sols = jnp.round(sols, 4)
        # print(ip * len(saveat.subs.ts), (ip * len(saveat.subs.ts)) + len(sols))
        # Add the solutions to the solution matrix at the correct index
        solution_matrix[ip * len_tsteps : (ip * len_tsteps) + len(sols)] = np.array(
            sols
        )
    # Convert the solution matrix to a dataframe
    # REmoving the InitCondNum column from the initial conditions
    if saveat.subs.ts is None:
        solution_matrix = pd.DataFrame(
            solution_matrix,
            columns=ic_columns[:-1].tolist()
            + ["Time", "SteadyStateFlag", "InitCondNum", "ParamNum"],
        )
        # Find number of steady state solutions
        # print(solution_matrix["SteadyStateFlag"].value_counts())
        # # Make the steady state flag, init cond and param num columns into integers
        # solution_matrix[["SteadyStateFlag", "InitCondNum", "ParamNum"]] = (
        #     solution_matrix[["SteadyStateFlag", "InitCondNum", "ParamNum"]].astype(int)
        # )
        # # Save the solution matrix as a parquet file
        # solution_matrix.to_parquet(
        #     f"{replicate_dir}/{topo_name}_steadystate_solutions.parquet", index=False
        # )
    else:
        solution_matrix = pd.DataFrame(
            solution_matrix,
            columns=ic_columns[:-1].tolist() + ["Time", "InitCondNum", "ParamNum"],
        )
        # # Make the init cond and param num columns into integers
        # solution_matrix[["InitCondNum", "ParamNum"]] = solution_matrix[
        #     ["InitCondNum", "ParamNum"]
        # ].astype(int)
        # Find number of last time points
        # print(solution_matrix["Time"].value_counts())
        # Save the solution matrix as a parquet file
        # solution_matrix.to_parquet(
        #     f"{replicate_dir}/{topo_name}_timeseries_solutions.parquet", index=False
        # )
    return solution_matrix


# Function to run all the replicate simulations for a given topo file
def run_all_replicates(
    topo_file,
    save_dir=".",
    t0=0.0,
    tmax=200.0,
    dt0=0.01,
    tsteps=None,
    rel_tol=1e-5,
    abs_tol=1e-6,
    max_steps=2048,
    batch_size=10000,
    normalize=True,
    discretize=True,
):
    """
    Run simulations for all replicates of the specified topo file. The initial conditions and parameters are loaded from the replicate folders. The directory structure is assumed to be the same as that generated by the gen_topo_param_files function, with the main directory with the topo file name which has the parameter range file the ODE system file and the replicate folders with the initial conditions and parameters dataframes.

    Parameters
    ----------
    topo_file : str
        Path to the topology file.
    save_dir : str, optional
        Directory where the replicate folders are saved. Defaults to ".".
    t0 : float, optional
        Initial time for the simulation. Defaults to 0.0.
    tmax : float, optional
        Maximum time for the simulation. Defaults to 100.0.
    dt0 : float, optional
        Initial time step for the simulation. Defaults to 0.1.
    tsteps : int, optional
        Number of time steps for the simulation. Defaults to None.
    rel_tol : float, optional
        Relative tolerance for the simulation. Defaults to 1e-5.
    abs_tol : float, optional
        Absolute tolerance for the simulation. Defaults to 1e-6.
    max_steps : int, optional
        Maximum number of steps for the simulation. Defaults to 2048.
    batch_size : int, optional
        Batch size for the simulation. Defaults to 1000.
    normalize : bool, optional
        Whether to normalise the solutions. Defaults to True.
    discretize : bool, optional
        Whether to discretize the solutions. Defaults to True.

    Returns
    -------
    None
        The results of the simulation are saved in the replicate folders in the specified directory.

    Note
    ----
    The results of the simulation are saved in the replicate folders in the specified directory. If the simulation is time series, the results are saved as `timeseries_solutions.parquet` and if the simulation is steady state, the results are saved as `steadystate_solutions.parquet`.

    Normalisation and discretisation of the solutions are optional. But to discretise the solutions, the normalisation is required.

    If the discretize flag is set to True, the solutions are discretized and the state counts are saved as `state_counts.csv`, this is only applicable for steady state simulations. If the discretize flag is set to False, the solutions are normalised but not discretized.

    Example
    --------
    Run the simulation for the specified topo file

        >>> run_all_replicates(topo_file, save_dir, t0, tmax, dt0, tsteps, rel_tol, abs_tol, max_steps, batch_size)
    """
    # Cheking if discretize is True and normalise is False, if so turn normalise to True
    if discretize and not normalize:
        normalize = True
    # Get the name of the topo file
    topo_name = topo_file.split("/")[-1].split(".")[0]
    # Get the list of replicate folders
    replicate_folders = sorted(
        [
            folder
            for folder in glob(f"{save_dir}/{topo_name}/*/")
            if os.path.basename(folder.rstrip("/")).isdigit()
        ],
    )
    # Loop through the replicate folders and run the simulation for each replicate
    for replicate_dir in replicate_folders:
        # Getting the base name of the replicate directory
        replicate_base = os.path.basename(replicate_dir.rstrip("/"))
        # Load the initial conditions and parameters dataframes
        init_cond_path = (
            f"{replicate_dir}/{topo_name}_init_conds_{replicate_base}.parquet"
        )
        params_path = f"{replicate_dir}/{topo_name}_params_{replicate_base}.parquet"
        # Read the initial conditions and parameters dataframes
        init_conds = pd.read_parquet(init_cond_path)
        params = pd.read_parquet(params_path)
        # Starting the timer
        start_time = time.time()
        # Run the simulation for the specified topo file and given initial conditions and parameters
        sol_df = topo_simulate(
            topo_file=topo_file,
            replicate_dir=replicate_dir,
            initial_conditions=init_conds,
            parameters=params,
            t0=t0,
            tmax=tmax,
            dt0=dt0,
            tsteps=tsteps,
            rel_tol=rel_tol,
            abs_tol=abs_tol,
            max_steps=max_steps,
            batch_size=batch_size,
        )
        # Ending the timer
        print(
            f"Time taken for replicate {replicate_base}: {time.time() - start_time}\n"
        )
        if discretize:
            # G/k normalise the solution dataframe
            sol_df, state_counts = gk_normalise_solutions(
                sol_df, params, discretize=True
            )
        elif normalize:
            sol_df = gk_normalise_solutions(sol_df, params, discretize=False)
        # Check if the time seires is given or not to name the solution file
        if tsteps is None:
            # Save the solution dataframe
            sol_df.to_parquet(
                f"{replicate_dir}/{topo_name}_steadystate_solutions_{replicate_base}.parquet"
            )
            if discretize:
                state_counts.to_csv(
                    f"{replicate_dir}/{topo_name}_steadystate_state_counts_{replicate_base}.csv",
                    index=False,
                    sep="\t",
                )
        else:
            # Read the solution dataframe
            sol_df.to_parquet(
                f"{replicate_dir}/{topo_name}_timeseries_solutions_{replicate_base}.parquet"
            )
        # # break  ##################################
    return None


# Function to g/k normalise the solution dataframe
def gk_normalise_solutions(sol_df, param_df, threshold=1.01, discretize=True):
    """
    Normalises the solutions in the solution dataframe using the maximum production rate (G) and degradation rate (k) parameters of the individual nodes in the parameter sets.

    Parameters
    ----------
    sol_df : pd.DataFrame
        DataFrame containing the solutions with a `ParamNum` column to join with param_df.
    param_df : pd.DataFrame
        DataFrame containing the parameters with `Prod_` and `Deg_` columns for each node.
    threshold : float, optional
        Threshold value for discretising the solutions. Default is 1.01.
    discretize : bool, optional
        Whether to discretise the solutions. Default is True.

    Returns
    -------
    pd.DataFrame
        DataFrame containing the normalised and discretised solutions.
    pd.DataFrame
        DataFrame containing the counts of each state in the normalised solutions.

    Example
    -------
    First, run the simulation then, normalise and discretise the solutions

        >>> sol_df, state_counts = gk_normalise_solutions(sol_df, params)

    Here, the `sol_df` is the solution dataframe and `params` is the parameter dataframe.
    The `state_counts` dataframe contains the counts of each state in the normalised solutions. The returned `sol_df` is the normalised and discretised solution dataframe.
    """
    # Get the node columns from the solution dataframe
    node_cols = [col.replace("Prod_", "") for col in param_df.columns if "Prod_" in col]
    # Get the production and degradation columns
    prod_cols = [f"Prod_{node}" for node in node_cols]
    deg_cols = [f"Deg_{node}" for node in node_cols]
    # Get the gk columns
    gk_cols = [f"gk_{node}" for node in node_cols]
    # Compute the gk values
    param_df[gk_cols] = param_df[prod_cols].values / param_df[deg_cols].values
    # Join the solution dataframe with the parameter dataframe on the 'ParamNum' column
    norm_df = sol_df.merge(param_df[gk_cols + ["ParamNum"]], on="ParamNum")
    # Divide the node columns by the gk columns
    norm_df[gk_cols] = norm_df[node_cols].values / norm_df[gk_cols].values
    if not discretize:
        # Select the node columns and discretise the solutions
        norm_df = pd.concat(
            [norm_df, discretise_solutions(norm_df[gk_cols], threshold)], axis=1
        )
    # Drop the gk columns
    norm_df.drop(columns=gk_cols, inplace=True)
    # print(norm_df)
    if discretize:
        # Get the State columns and return the state counts
        state_counts = norm_df["State"].value_counts()
        # Make the State column as a column
        state_counts = state_counts.reset_index()
        # Rename the columns
        state_counts.columns = ["State", "Count"]
        # Return the normalised and discretised solution dataframe
        return norm_df, state_counts
    else:
        return norm_df


# Function to process the output to discretise the data
def discretise_solutions(norm_df, threshold=1.01):
    """
    Discretises the solutions in a g/k normalized DataFrame based on histogram peaks and minima.

    Parameters
    ----------
    norm_df : pd.DataFrame
        DataFrame containing normalized values to be discretised.
    threshold : float, optional
        A hard threshold value to clip the values in the DataFrame. Default is 1.01. If the parameter sets are in such a way that the maximum possible expression of the node is not production/degradation, then the threshold value needs to be adjusted accordingly.

    Returns
    -------
    pd.Series
        A Series containing the discrete state labels for each row in the input DataFrame.

    Raises
    ------
    ValueError
        If any value in the DataFrame exceeds the specified threshold.

    Example
    -------
    Given a normalized DataFrame ```norm_df```, discretise the values

        >>> lvl_df = discretise_solutions(norm_df)

    The normalized solution DataFrame contains values of the nodes between 0 (lowest) and 1 (highest).
    The returned `lvl_df` will have discrete state labels for each row in the input DataFrame.

    """
    # Flatten the numeric part of the dataframe (columns 4 onwards)
    flat = norm_df.values.flatten()

    # Add dummy values to ensure boundary peaks are detected
    dummy_low = np.full(10, -0.1)
    dummy_high = np.full(10, 1.1)
    data_with_dummy = np.concatenate([dummy_low, flat, dummy_high])

    # Compute histogram over 120 bins
    flat_hist, bin_edges = np.histogram(data_with_dummy, bins=120)

    # Define threshold for peak and minima detection (1% of data length)
    threshold = int(len(flat) * 0.01)

    # Detect peaks in the histogram
    peaks, _ = find_peaks(
        flat_hist,
        height=threshold,
        threshold=threshold,
        prominence=threshold,
        distance=int(len(bin_edges) * 0.1),
    )
    maxima_bins = bin_edges[peaks]

    # Detect minima by inverting the histogram (skip first and last bin)
    minima_indices, _ = find_peaks(
        np.max(flat_hist) - flat_hist[1:-1], height=np.max(flat_hist) * 0.99
    )
    # Adjust indices (because we skipped the first bin)
    minima_indices = [idx + 1 for idx in minima_indices]

    # Initialize minima bins with the boundaries 0.0 and 1.0
    minima_bins = [0.0, 1.0]

    # For multiple peaks, find a representative (median) minimum between each adjacent pair
    if len(maxima_bins) > 1:
        for i in range(1, len(maxima_bins)):
            # Candidate minima between two successive peaks
            candidates = [
                bin_edges[idx]
                for idx in minima_indices
                if maxima_bins[i - 1] <= bin_edges[idx] <= maxima_bins[i]
            ]
            if candidates:
                median_candidate = candidates[len(candidates) // 2]
                minima_bins.append(median_candidate)
    # If only one peak exists, take the median minimum if available
    elif minima_indices:
        median_candidate = bin_edges[minima_indices[len(minima_indices) // 2]]
        minima_bins.append(median_candidate)

    # Ensure the bin edges are sorted
    minima_bins = np.sort(minima_bins)

    # Clip the values to 1.0 as due to small numerical errors, some values may be slightly above 1.0
    norm_df = norm_df.mask((norm_df > 1) & (norm_df < threshold), 1.0)

    # If any value is found to be higher than the hard threshold, raise an error
    if (norm_df > threshold).any().any():
        raise ValueError(
            f"Some values exceed the hard threshold of {threshold}, check your input data."
        )

    # Use vectorized binning to assign each value to a discrete level
    # The number of levels is len(minima_bins)-1 (each interval defines one level)
    lvl_df = norm_df.apply(
        lambda col: pd.cut(
            col,
            bins=minima_bins,
            labels=range(len(minima_bins) - 1),
            include_lowest=True,
        )
    )
    lvl_df = lvl_df.add_prefix("Lvl_")

    # Create a 'State' column by concatenating the discrete levels as strings
    lvl_df["State"] = "'" + lvl_df.astype(str).apply("".join, axis=1) + "'"
    # print(lvl_df)
    # Get the value counts of the states
    # print(lvl_df["State"].value_counts())
    return lvl_df["State"]


if __name__ == "__main__":
    pass
    # # Specify the number of cores to use
    # numCores = 15
    # print(f"Number of cores: {numCores}")
    # # Topo file directory
    # topo_dir = "TOPOS"
    # # Specify the root folder where the generated parameter files and then the simulation files will be saved
    # sim_save_dir = "SimulResults"
    # # Make the directories to store the results
    # os.makedirs(sim_save_dir, exist_ok=True)
    # # Get the list of all the topo files
    # topo_files = sorted(glob(f"{topo_dir}/*.topo"))
    # # Remove topo files with 50N
    # topo_files = [topo_file for topo_file in topo_files if "50N" not in topo_file]
    # print(f"Number of topo files: {len(topo_files)}")
    # # Specify the number of replicates required
    # num_replicates = 3
    # # Specify the number of parameters required
    # num_params = 10000
    # # Specify the number of initial conditions required
    # num_init_conds = 100
    # # Print the number of replicates, parameters and initial conditions
    # print(f"Number of replicates: {num_replicates}")
    # print(f"Number of parameters: {num_params}")
    # print(f"Number of initial conditions: {num_init_conds}\n")
    # # Start the pool of worker processes
    # pool = Pool(int(numCores))
    # # Parllelise the generation of the parameter and inital condition files
    # print("Generating Parameter and Initial Condition files...")
    # pool.starmap(
    #     gen_topo_param_files,
    #     [
    #         (
    #             topo_file,
    #             sim_save_dir,
    #             # sim_ode_dir,
    #             num_replicates,
    #             num_params,
    #             num_init_conds,
    #         )
    #         for topo_file in topo_files
    #     ],
    # )
    # print("Parameter and Initial Condition files generated.\n")
    # # Close the pool of workers
    # pool.close()
    # # Loop through the topo files and load the ODE system
    # for topo_file in topo_files:
    #     # Generate the parameter files, intial condition files and the directory structure
    #     gen_topo_param_files(
    #         topo_file,
    #         sim_save_dir,
    #         num_replicates,
    #         num_params,
    #         num_init_conds,
    #         sampling_method="Sobol",
    #     )
    #     # Call the function to run the simulation for the specified topo file - Time series
    #     run_all_replicates(
    #         topo_file,
    #         sim_save_dir,
    #         tsteps=jnp.array([25.0, 75.0, 100.0]),
    #         max_steps=2048,
    #     )
    #     # Call the function to run the simulation for the specified topo file - Steady state
    #     run_all_replicates(
    #         topo_file,
    #         sim_save_dir,
    #     )
    #     # break



# Positive Shifted Hill function
def psH(nod, fld, thr, hill):
    """
    Positive Shifted Hill function.

    Parameters
    ----------
    nod : float
        The node expression value.
    fld : float
        The fold change.
    thr : float
        The half-maximal threshold value.
    hill : float
        The hill coefficient.

    Returns
    -------
    float
        The value of the Positive Shifted Hill function.
    """
    return (fld + (1 - fld) * (1 / (1 + (nod / thr) ** hill))) / fld


# Negative Shifted Hill function
def nsH(nod, fld, thr, hill):
    """
    Negative Shifted Hill function.

    Parameters
    ----------
    nod : float
        The node expression value.
    fld : float
        The fold change.
    thr : float
        The half-maximal threshold value.
    hill : float
        The hill coefficient.

    Returns
    -------
    float
        The value of the Negative Shifted Hill function.
    """
    return fld + (1 - fld) * (1 / (1 + (nod / thr) ** hill))



import os
import shutil
import glob
import argparse
import pandas as pd
from grins.generate_params import parse_topos, gen_param_df, get_param_range_df, gen_init_cond, get_thr_ranges
from grins.gen_diffrax_ode import gen_diffrax_odesys
from grins.sim_run import get_steady_states, ODE_topo

def main():
    parser = argparse.ArgumentParser(\
        prog='RACIPE',
        description='Run simulation of GRN-ODE model for ensemble of parameters')
    parser.add_argument('topo', type=str, help='topo file name', default='all', nargs='?')
    parser.add_argument('--topodir', type=str, help='topo file directory', default='TOPOS')
    parser.add_argument('--outdir', type=str, help='simulation directory', default='SimResults')
    parser.add_argument('--num_paras', type=int, help='number of parameters', default=10000)
    parser.add_argument('--num_inits', type=int, help='number of initial conditions', default=1000)
    parser.add_argument('--num_reps', type=int, help='number of replicates', default=3)
    parser.add_argument('--num_cores', type=int, help='number of cores', default=0)
    parser.add_argument('--genprs', action='store_true', help='generate parameter range file, no simulation')
    parser.add_argument('--prsfile', type=str, help='parameter range file name', default=None)
    parser.add_argument('--no_calc_thrs', action='store_true', help='Do not calculate thresholds, used with prsfile')
    parser.add_argument('--sampling', type=str, help='sampling method. Choices: sobol, uni, loguni, latin_hc', default='sobol')
    # parser.add_argument('--gencfg', type=bool, help='generate config file only, no simulation', default=False)
    # parser.add_argument('--cfgfile', type=str, help='config file name', default=None)
    args = parser.parse_args()
    # if no topo file is provided, use iterate over all the topo files
    topos = sorted(glob.glob(f"{args.topodir}/*.topo")) if args.topo == 'all' else [args.topo]
    for tpfl in topos:
        # Get the topo name
        topo_name = tpfl.split("/")[-1].split(".")[0]
        # Create the directory
        os.makedirs(f"{args.outdir}/{topo_name}", exist_ok=True)
        # Parse the topology file
        topo_df = parse_topos(tpfl)
        # Copy topofile to outdir
        shutil.copy(tpfl, f"{args.outdir}/{topo_name}")
        # Switch cases for cfg, prs
        if args.genprs:
            gen_prs(topo_name, topo_df, args)
        else:
            default(topo_name, topo_df, args)

def gen_prs(topo_name, topo_df, args):
    print(f'Generating Parameter range for {topo_name}')
    # Generate the parameter names
    prange_df = get_param_range_df(topo_df, sampling=args.sampling)
    prange_df.to_csv(f"{args.outdir}/{topo_name}/{topo_name}.prs", index=True, sep='\t')
    return prange_df
    
def default(topo_name, topo_df, args):
    # Generate the parameter names
    if args.prsfile:
        prange_df = pd.read_csv(args.prsfile, sep='\t', index_col=0)
        if not args.no_calc_thrs:
            get_thr_ranges(topo_df, prange_df)
    else:
        prange_df = gen_prs(topo_name, topo_df, args)
    # Generating parameters for the topology
    # Generate the parameter dataframe
    print(f'Generating Parameters for {topo_name}')
    param_df = gen_param_df(prange_df, num_paras=args.num_paras, sampling=args.sampling)
    param_df.to_csv(f"{args.outdir}/{topo_name}/{topo_name}_params.dat", sep='\t', index=True)
    # Generate the initial conditions dataframe
    init_df = gen_init_cond(topo_df, num_init_conds=args.num_inits, sampling=args.sampling)
    init_df.to_csv(f"{args.outdir}/{topo_name}/{topo_name}_init_conds.csv", index=True)
    print(f'Generating ODE for {topo_name}')
    # Generate the diffrax ode system
    gen_diffrax_odesys(
        topo_df, topo_name, save_dir=f"{args.outdir}/{topo_name}"
    )
    term = ODE_topo(topo_name, args.outdir)
    # Get the steady states
    print(f'Calculating steady states for {topo_name}')
    sol_df = get_steady_states(init_df, param_df, term)
    sol_df.to_parquet(f"{args.outdir}/{topo_name}/{topo_name}_sol.parquet", index=True)

if __name__ == '__main__':
    main()


import glob
from itertools import chain, cycle
import numpy as np
import pandas as pd


# Function to find all the state variables in the logical boolean model
def find_state_variables(bool_lines, bool_logic_ops):
    """
    Processes a list of boolean logic expressions to identify state variables and their indices.

    Args:
        bool_lines (list of str): List of boolean logic expressions as strings.
        bool_logic_ops (dict): Dictionary of boolean logic operations.

    Returns:
        tuple: A tuple containing:
            - dict: A dictionary mapping each unique state variable to its index.
            - list: A list of state variable nodes (left-hand side of the expressions).

    Raises:
        ValueError: If a line does not have '=' as the second element.
    """
    # Replace all the parentheses with white spaces
    bool_lines = [line.replace("(", " ").replace(")", " ") for line in bool_lines]
    # Split the lines at white spaces
    bool_lines = [line.split() for line in bool_lines]
    # Make a placeholder list of the nodes for which logical expressions are defined (LHS of the expression)
    statevar_nodes = []
    # Loop through all the lines
    for line in bool_lines:
        # Checking if = is the second element in the line
        if "=" in line[1]:
            # Append the first element of the line to the logicexpr_nodes list
            statevar_nodes.append(line[0])
        else:
            # Raise an error if the line does not have = as the second element
            # Print the index of the line
            raise ValueError(
                f"Logical expression not defined properly ({bool_lines.index(line)})"
            )
    # Get the unique state variables
    all_boolnodes = set(list(chain(*bool_lines)))
    # Remove all boolean operations and "=" from the state variables
    all_boolnodes = all_boolnodes - set(bool_logic_ops.values()) - set("=")
    # Sort the state variables - ensures the order of the state variables is consistent across all runs
    all_boolnodes = sorted(list(all_boolnodes))
    # Make a map of the bool nodes and thier index in the list
    all_boolnodes = {node: i for i, node in enumerate(all_boolnodes)}
    # Return all the state variables and all boolean nodes
    return all_boolnodes, statevar_nodes


# Function to replace all the node names with their index in the list for a given line
def replace_node_names(line, all_boolnodes):
    """
    Replaces node names in a given line with their corresponding indices from a dictionary.

    Args:
        line (str): The input string containing node names to be replaced.
        all_boolnodes (dict): A dictionary where keys are node names and values are their corresponding indices.

    Returns:
        str: The modified line with node names replaced by their indices in the format 'bnode[index]'.
    """
    # Loop through all the nodes
    for key in all_boolnodes.keys():
        # Check if the node is present in the line
        if key in line:
            # Replace the node with its index in the list
            line = line.replace(key, f"bnode[{all_boolnodes[key]}]")
    # Return the line with the node names replaced with their index
    return line


# Function to the read the logical boolean model file
def read_logical_boolmodel(filename, bool_logic_ops):
    """
    Reads and parses a logical boolean model from a file, replacing custom logical operators
    with Python logical operators, and returns the state variables and logical expressions.

    Args:
        filename (str): The path to the file containing the logical boolean model.
        bool_logic_ops (dict): A dictionary mapping custom logical operators to Python logical operators.

    Returns:
        tuple: A tuple containing:
            - all_boolnodes (dict): A dictionary mapping state variable names to their indices.
            - lines (list): A list of logical expressions with state variables replaced by their indices.
    """
    # Open the file and parse it
    with open(filename, "r") as f:
        # Read all the lines of the file
        lines = f.readlines()
        # Strip the lines of any leading or trailing whitespaces
        lines = [line.strip() for line in lines]
        # Strip the lines of newline characters
        lines = [line.replace("\n", "") for line in lines]
        # Replace all other whitespaces with a single whitespace
        lines = [" ".join(line.split()) for line in lines]
        # Replace the logical operators with the python logical operators
        for key, value in bool_logic_ops.items():
            lines = [line.replace(key, value) for line in lines]
        # Find all the state variables in the logical boolean model
        all_boolnodes, statevar_nodes = find_state_variables(lines, bool_logic_ops)
        # Replace all the node names with their index in the list
        lines = [replace_node_names(line, all_boolnodes) for line in lines]
    # Some nodes do not have logical expressions defined (Input nodes)
    # Adding thier expression values as Node = Node
    for node in all_boolnodes.keys():
        if node not in statevar_nodes:
            lines.append(f"bnode[{all_boolnodes[node]}] = bnode[{all_boolnodes[node]}]")
    # Sort the lines according to the state variables index
    lines = sorted(
        lines, key=lambda x: int(x.split("=")[0].replace("bnode[", "").replace("]", ""))
    )
    # Remove all LHS of the logical expressions as we already have ordered them according to the state variables
    # This will make it easier for tracking and updating the state at each step (inplace update can be avoided)
    lines = [line.split("=")[1].strip() for line in lines]
    # Return the state variables and the logical expressions
    return all_boolnodes, lines


# Function to convert the simulation states to a pandas dataframe
def generate_solution_df(simulation_states, node_list):
    """
    Convert the simulation states to a pandas dataframe.

    Parameters:
    simulation_states (list): The list of states of the nodes.
    node_list (list): The list of nodes in the network.

    Returns:
    pandas.DataFrame: The converted dataframe.
    """
    # Convert the simulation states to a pandas dataframe
    simulation_states = pd.DataFrame(simulation_states, columns=["Step"] + node_list)
    # Replace all the True values with 1 and False values with 0 - only applicable for the initial and steady states as they become string and not boolean
    simulation_states[node_list] = simulation_states[node_list].replace(
        {"True": 1, "False": 0}
    )
    # Convert the state of the nodes from boolean to integer
    simulation_states[node_list] = simulation_states[node_list].astype(int)
    # Check if the last step is the steady state if not make it as UnSteady
    # Getting the last row of the simulation states and checking if the state labelled as "Steady"
    if simulation_states["Step"].iloc[-1] != "SteadyState":
        simulation_states.at[simulation_states.index[-1], "Step"] = "UnSteady"
    # Return the simulation states dataframe
    return simulation_states


# Function to generate the final state dictionary
def generate_final_state_dict(simulation_states, all_boolnodes):
    """
    Generate the final state dictionary from the simulation states.

    Parameters:
    simulation_states (list): The list of states of the nodes.
    all_boolnodes (dict): A dictionary mapping state variable names to their indices.

    Returns:
    dict: The final state dictionary.
    """
    # Get the last simulation state
    # Replace the True and False values with 1 and 0
    last_state = [1 if node == "True" else 0 for node in simulation_states[-1][1:]]
    # Make a dictionary of the final state
    last_state = dict(zip(list(all_boolnodes.keys()), last_state))
    # Check if the last step is the steady state if not make it as UnSteady
    if simulation_states[-1][0] != "SteadyState":
        last_state["StateType"] = "UnSteady"
    else:
        last_state["StateType"] = "SteadyState"
    # Add time step to the dictionary
    last_state["Step"] = len(simulation_states) - 1
    # return the dictionary
    return last_state


# Function to update the node state based on the logical expression
def update_node_state(node_index, logical_expresion, bnode):
    """
    Update the state of a node based on a logical expression.

    Args:
        node_index (int): The index of the node to update.
        logical_expresion (list): A list of logical expressions.
        bnode (list): A list representing the current state of nodes.

    Returns:
        list: The updated state of the nodes.
    """
    # Update the node state based on the logical expression
    new_node_state = eval(logical_expresion[node_index])
    # Update the state of the node in bnode
    bnode[node_index] = new_node_state
    # Return the updated state of the node
    return bnode


# Function run the logical boolean model
def run_logicalbool_sync(
    logic_file,
    max_steps=None,
    inital_state=None,
    bool_logic_ops={"AND": "and", "OR": "or", "NOT": "not"},
    time_series=True,
):
    """
    Simulates the synchronous update of a logical boolean model.

    Parameters:
    logic_file (str): Path to the file containing the logical boolean model.
    max_steps (int, optional): Maximum number of steps for the simulation. If not provided, defaults to 100 times the number of state variables.
    inital_state (list of bool, optional): Initial state of the state variables. If not provided, the state variables are randomly initialized.
    bool_logic_ops (dict, optional): Dictionary mapping logical operators to their corresponding string representations. Defaults to {"AND": "and", "OR": "or", "NOT": "not"}.
    time_series (bool, optional): If True, returns the simulation states as a time series DataFrame. If False, returns the final state as a dictionary. Defaults to True.

    Returns:
    dict or pandas.DataFrame: If time_series is False, returns a dictionary with the final state of the state variables, state type and the timestep of the final state. If time_series is True, returns a DataFrame with the simulation states over time.
    """
    # Read the logical boolean model
    all_boolnodes, logical_expressions = read_logical_boolmodel(
        logic_file, bool_logic_ops
    )
    # Get the number of state variables
    num_state_vars = len(all_boolnodes)
    # If the max_steps is not provided, set it to be equal to 100*number of state variables
    if max_steps is None:
        max_steps = 100 * num_state_vars
    # else:
    #     max_steps = max_steps * num_state_vars
    # If the initial state is not provided, randomly initialize the state variables
    # The values will be true or false values
    if inital_state is None:
        inital_state = np.random.choice([True, False], num_state_vars).tolist()
    # Store the states and the step numbers as a list of lists
    simulation_states = []
    # Set the initial state of the state variables
    bnode = inital_state
    # Add the initial state to the simulation states
    simulation_states.append(["InitialState"] + bnode)
    # print("Initial State: ", bnode)
    # Loop through the max_steps
    for i in range(1, max_steps):
        # Map the excec function on the logical expressions list
        new_bnode = list(map(eval, logical_expressions))
        # Check if new state is same as the previous state
        if np.array_equal(new_bnode, bnode):
            # Update the new state as the final state and break the loop
            bnode = new_bnode
            # Append the final state to the simulation states
            simulation_states.append(["SteadyState"] + bnode)
            # print("Final State: ", bnode)
            break
        # Update the state of the state variables
        bnode = new_bnode
        # Append the state of the state variables to the simulation states
        simulation_states.append([i] + bnode)
    if time_series:
        # Return the simulation states dataframe
        return generate_solution_df(simulation_states, list(all_boolnodes.keys()))
    else:
        # Return the final state dictionary
        return generate_final_state_dict(simulation_states, all_boolnodes)


# # Function to check is the previous n states are the same as the current state
# def check_convergence(current_state, simulation_states, min_steps):
#     # Get the last n states of the simulation states
#     last_n_states = simulation_states[:, 1:][-min_steps:]
#     # Check if the last n states are the same and if its the same as the current state
#     if (last_n_states == current_state).all():
#         return True
#     else:
#         return False


# Function to check convergence using one step of sync update to see if the same state is reached
def check_convergence_sync_step(bnode, logical_expressions):
    # Get the state of the nodes after one step of sync update
    new_state = list(map(eval, logical_expressions))
    # Check if bnnode is the same as the new state
    if np.array_equal(bnode, new_state):
        return True
    else:
        return False


# Function to run the logical boolean model in an asynchronous mode
def run_logicalbool_async(
    logic_file,
    max_steps=None,
    inital_state=None,
    update_order=None,
    min_steps=None,
    bool_logic_ops={"AND": "and", "OR": "or", "NOT": "not"},
    time_series=True,
):
    """
    Simulates the asynchronous update of a logical boolean model.

    Parameters:
    logic_file (str): Path to the file containing the logical boolean model.
    max_steps (int, optional): Maximum number of steps for the simulation. Defaults to 100 times the number of state variables.
    inital_state (list of bool, optional): Initial state of the state system. If None, the state variables are randomly initialized. Defaults to None.
    update_order (list of str or str, optional): Order in which the nodes are updated. If None, nodes are randomly sampled each step. If "random_cyclic", nodes are updated in a random cyclic order. Defaults to None.
    min_steps (int, optional): Minimum number of steps for the simulation. Defaults to the number of state variables.
    bool_logic_ops (dict, optional): Dictionary mapping logical operators to their string representations. Defaults to {"AND": "and", "OR": "or", "NOT": "not"}.
    time_series (bool, optional): If True, returns the simulation states as a time series DataFrame. If False, returns the final state, state type and the timestep of the final state as a Dictionary. Defaults to True.

    Returns:
    pandas.DataFrame or dict: If time_series is True, returns a DataFrame containing the simulation states over time. If time_series is False, returns a dictionary containing the final state of the state variables.
    """
    # Read the logical boolean model
    all_boolnodes, logical_expressions = read_logical_boolmodel(
        logic_file, bool_logic_ops
    )
    # Get the number of state variables
    num_state_vars = len(all_boolnodes)
    # If the min_steps is not provided, set it to be equal to number of state variables
    if min_steps is None:
        min_steps = num_state_vars
    # If the max_steps is not provided, set it to be equal to 100*number of state variables
    if max_steps is None:
        max_steps = 100 * num_state_vars
    # else:
    #     max_steps = max_steps * num_state_vars
    # If the initial state is not provided, randomly initialize the state variables
    # The values will be true or false values
    if inital_state is None:
        inital_state = np.random.choice([True, False], num_state_vars).tolist()
    # Set the initial state of the state variables
    bnode = inital_state
    # Initialize the simulation states matrix with the initial state
    simulation_states = np.array([["InitialState"] + bnode])
    # If node cycle is None, randomly sample the nodes from the state variables
    if update_order is None:
        update_order = np.random.choice(range(0, len(all_boolnodes)), max_steps)
    elif update_order == "random_cyclic":
        # Create a order of the ndoes by randomly shuffling the node list
        update_order = np.random.permutation(list(all_boolnodes.keys()))
        # Map the node names to their index in the list
        update_order = [all_boolnodes[node] for node in update_order]
    else:
        # Map the node names to their index in the list
        update_order = [all_boolnodes[node] for node in update_order]
    # Make the node order cyclic
    update_order = cycle(update_order)
    # Initialize the step counter
    step = 1
    # Loop thorugh the cycle of the node order
    for node_to_update in update_order:
        # Break the loop if the step counter exceeds the max_steps
        if step >= max_steps:
            break
        # Update the state of the node
        new_bnode = update_node_state(node_to_update, logical_expressions, bnode)
        # Check if new state is same as the previous state
        # Also check is the current state is the same as the previous
        if step > (min_steps + 1) and np.array_equal(new_bnode, bnode):
            # Now check if synchrnous update will reach the same state
            if check_convergence_sync_step(bnode, logical_expressions):
                # Update the new state as the final state and break the loop
                bnode = new_bnode
                # Append the final state to the simulation states
                simulation_states = np.append(
                    simulation_states, [["SteadyState"] + bnode], axis=0
                )
                break
        # Update the state of the state variables
        bnode = new_bnode
        # Append the state of the state variables to the simulation states
        simulation_states = np.append(simulation_states, [[step] + bnode], axis=0)
        # Increment the step counter
        step += 1
    if time_series:
        simulation_states = generate_solution_df(
            simulation_states, list(all_boolnodes.keys())
        )
        # Return the simulation states
        return simulation_states
    else:
        # Generate the final state dictionary
        final_state = generate_final_state_dict(simulation_states, all_boolnodes)
        # Return the final state dictionary
        return final_state


if __name__ == "__main__":
    # Specifying the folder path
    boolmodel_path = "../Logical_TOPOS/"
    # Get a list of all models in the folder
    model_list = sorted(glob.glob(boolmodel_path + "*.txt"))
    # Loop through all the models
    for model_file in model_list[:1]:
        print(f"Model: {model_file}")
        # Run the logical boolean model
        for i in range(3):
            print(f"Run {i}")
            solu_df = run_logicalbool_async(
                model_file,
                # update_order="random_cyclic",
                time_series=False,
            )
            print(solu_df)



import pandas as pd
import numpy as np
from glob import glob
from scipy.stats import qmc, truncnorm
import itertools as it
import warnings
from grins.reg_funcs import psH, nsH
from typing import Tuple, List, Union, Optional

# Suppress the specific warning
warnings.filterwarnings(
    "ignore",
    message="The balance properties of Sobol' points require n to be a power of 2",
    category=UserWarning,
)


def parse_topos(topofile: str, save_cleaned: bool = False) -> pd.DataFrame:
    """
    Parse and cleans the given topofile and return the dataframe. It is expeced that the topo files is a tab-separated file with three columns: Source, Target, and Type. White spaces can also be used to separate the columns.

    For nodes that are not alphanumeric, the function will replace the non-alphanumeric characters with an underscore and prepend "Node_" if the node name does not start with an alphabet. The cleaned topology file will be saved if the save_cleaned flag is set to True.

    Parameters
    ----------
    topofile : str
        The path to the topofile.
    save_cleaned : bool, optional
        If True, save the cleaned topology file. Defaults to False.

    Returns
    -------
    topo_df : pd.DataFrame
        The parsed dataframe.
    """
    topo_df = pd.read_csv(topofile, sep=r"\s+")
    if topo_df.shape[1] != 3:
        raise ValueError(
            "The topology file should have three columns: Source, Target, and Type."
        )
    if not all(col in topo_df.columns for col in ["Source", "Target", "Type"]):
        raise ValueError(
            "The topology file should have the columns: Source, Target, and Type."
        )
    topo_df = topo_df[["Source", "Target", "Type"]]
    # Clean up node names: replace non-alphanumerics and prepend "Node_" if needed.
    topo_df["Source"] = (
        topo_df["Source"]
        .str.replace(r"\W", "_", regex=True)
        .apply(lambda x: f"Node_{x}" if not x[0].isalpha() else x)
    )
    topo_df["Target"] = (
        topo_df["Target"]
        .str.replace(r"\W", "_", regex=True)
        .apply(lambda x: f"Node_{x}" if not x[0].isalpha() else x)
    )
    if topo_df["Type"].nunique() > 2:
        raise ValueError(f"Check the topo file: {topofile}")
    if save_cleaned:
        topo_df.to_csv(
            topofile.replace(".topo", "_cleaned.topo"), sep="\t", index=False
        )
    return topo_df


def _get_regtype(sn: str, tn: str, topo_df: pd.DataFrame) -> str:
    """
    Get the type of regulation for a given source and target node.

    Parameters
    ----------
    sn : str
        The source node.
    tn : str
        The target node.
    topo_df : pd.DataFrame
        The DataFrame containing the topology information.

    Returns
    -------
    str
        The type of regulation, either "ActFld_{sn}_{tn}" for activation or "InhFld_{sn}_{tn}" for inhibition.
    """
    reg_type = topo_df[(topo_df["Source"] == sn) & (topo_df["Target"] == tn)][
        "Type"
    ].iloc[0]
    return f"ActFld_{sn}_{tn}" if reg_type == 1 else f"InhFld_{sn}_{tn}"


def gen_param_names(topo_df: pd.DataFrame) -> Tuple[List[str], List[str], List[str]]:
    """
    Generate parameter names based on the given topology dataframe.

    Parameters
    ----------
    topo_df : pd.DataFrame
        The topology dataframe containing the information about the nodes and edges.

    Returns
    -------
    tuple
        A tuple containing the parameter names, unique target node names, and unique source node names.
    """
    target_nodes = list(topo_df["Target"].unique())
    source_nodes = list(topo_df["Source"].unique())
    unique_nodes = sorted(set(source_nodes + target_nodes))
    param_names = [f"Prod_{n}" for n in unique_nodes] + [
        f"Deg_{n}" for n in unique_nodes
    ]
    for tn in unique_nodes:
        sources = topo_df[topo_df["Target"] == tn]["Source"].sort_values()
        for sn, p in it.product(sources, ["Fld", "Thr", "Hill"]):
            param_names.append(
                f"{p}_{sn}_{tn}" if p != "Fld" else _get_regtype(sn, tn, topo_df)
            )
    return param_names, target_nodes, source_nodes


def scale_array(
    arr: np.ndarray, min_val: float, max_val: float, round_int: bool = False
) -> np.ndarray:
    """
    Scale the array values to the range [min_val, max_val]. If round_int is True, np.ceil the results
    and np.clip to the range.

    Parameters
    ----------
    arr : np.ndarray
        The input array to be scaled.
    min_val : float
        The minimum value of the scaled array.
    max_val : float
        The maximum value of the scaled array.
    round_int : bool, optional
        If True, ceil the results and clip to the range. Defaults to False.

    Returns
    -------
    np.ndarray
        The scaled array with values in the range [min_val, max_val].
    """
    scaled = min_val + (max_val - min_val) * (arr - arr.min()) / (arr.max() - arr.min())
    if round_int:
        scaled = np.ceil(scaled)
        scaled = np.clip(scaled, min_val + 1, max_val)
    return scaled


def _gen_sobol_seq(
    dimensions: int, num_points: int, optimise: bool = False
) -> np.ndarray:
    """
    Generate a Sobol sequence.

    Parameters:
        dimensions (int): The number of dimensions for the Sobol sequence.
        num_points (int): The number of points to generate in the sequence.
        optimise (bool, optional): If True, apply Lloyd's optimization to the sequence. Default is False.

    Returns:
        np.ndarray: An array containing the generated Sobol sequence.
    """
    sampler = qmc.Sobol(
        d=dimensions, scramble=True, optimization="lloyd" if optimise else None
    )
    return sampler.random(num_points)


def _gen_uniform_seq(dimension: int, num_points: int) -> np.ndarray:
    """
    Generate a sequence of uniformly distributed random points.

    Parameters:
        dimension (int): The number of dimensions for each point.
        num_points (int): The number of points to generate.

    Returns:
        np.ndarray: A 2D array of shape (num_points, dimension) containing the generated points.
    """
    return np.random.uniform(low=0, high=1, size=(num_points, dimension))


def _gen_loguniform_seq(dimension: int, num_points: int) -> np.ndarray:
    """
    Generate a sequence of points sampled from a log-uniform distribution.

    Parameters:
        dimension (int): The dimensionality of the points.
        num_points (int): The number of points to generate.

    Returns:
        np.ndarray: An array of shape (num_points, dimension) containing the generated points.
    """
    return np.exp(_gen_uniform_seq(dimension, num_points))


def _gen_latin_hypercube(
    dimension: int, num_points: int, optimise: bool = False
) -> np.ndarray:
    """
    Generate a Latin Hypercube sample.

    Parameters:
        dimension (int): The number of dimensions for the sample.
        num_points (int): The number of points to generate.
        optimise (bool, optional): Whether to use Lloyd's algorithm for optimization. Defaults to False.

    Returns:
        np.ndarray: A numpy array containing the generated Latin Hypercube sample.
    """
    sampler = qmc.LatinHypercube(
        d=dimension, scramble=True, optimization="lloyd" if optimise else None
    )
    return sampler.random(num_points)


def _gen_normal(dimension: int, num_points: int, std_dev: float = 1) -> np.ndarray:
    """
    Generate a set of points from a truncated normal distribution.

    This function generates `num_points` points in `dimension`-dimensional space,
    where each point is drawn from a truncated normal distribution with a mean of 0
    and a standard deviation of `std_dev`. The distribution is truncated to avoid
    extreme values, limiting the range to ±3 standard deviations.

    Parameters:
        dimension (int): The number of dimensions for each point.
        num_points (int): The number of points to generate.
        std_dev (float, optional): The standard deviation of the normal distribution. Default is 1.

    Returns:
        np.ndarray: A NumPy array of shape (num_points, dimension) containing the generated points.
    """
    # Use truncated normal to avoid extreme values (±3 std deviations)
    lower, upper = -3, 3
    return truncnorm.rvs(
        a=lower, b=upper, loc=0, scale=std_dev, size=(num_points, dimension)
    )


def _gen_lognormal(dimension: int, num_points: int, std_dev: float = 1) -> np.ndarray:
    """
    Generate a log-normal distribution.

    Parameters:
        dimension (int): The number of dimensions for each point.
        num_points (int): The number of points to generate.
        std_dev (float, optional): The standard deviation of the underlying normal distribution. Default is 1.

    Returns:
        np.ndarray: An array of shape (num_points, dimension) containing the generated log-normal points.
    """
    return np.exp(_gen_normal(dimension, num_points, std_dev))


def sample_distribution(
    method: str,
    dimension: int,
    num_points: int,
    std_dev: Optional[float] = None,
    optimise: bool = False,
) -> np.ndarray:
    """
    Generates a sample distribution based on the specified method.

    Parameters
    ----------
    method : str
        The sampling method to use. Options are "Sobol", "LHS", "Uniform", "LogUniform", "Normal", "LogNormal".
    dimension : int
        The number of dimensions for the sample points.
    num_points : int
        The number of sample points to generate.
    std_dev : Optional[float], optional
        The standard deviation for the "Normal" and "LogNormal" distributions. Defaults to 1.0 if not provided.
    optimise : bool, optional
        Whether to optimise the sampling process. Applicable for "Sobol" and "LHS" methods.

    Returns
    -------
    np.ndarray
        An array of sample points generated according to the specified method.

    Raises
    ------
    ValueError
        If an unknown sampling method is specified.
    """
    if method == "Sobol":
        dist_arr = _gen_sobol_seq(dimension, num_points, optimise)
        # return _gen_sobol_seq(dimension, num_points, optimise)
    elif method == "LHS":
        dist_arr = _gen_latin_hypercube(dimension, num_points, optimise)
        # return _gen_latin_hypercube(dimension, num_points, optimise)
    elif method == "Uniform":
        dist_arr = _gen_uniform_seq(dimension, num_points)
        # return _gen_uniform_seq(dimension, num_points)
    elif method == "LogUniform":
        dist_arr = _gen_loguniform_seq(dimension, num_points)
        # return _gen_loguniform_seq(dimension, num_points)
    elif method == "Normal":
        dist_arr = _gen_normal(
            dimension, num_points, std_dev if std_dev is not None else 1.0
        )
        # return _gen_normal(
        #     dimension, num_points, std_dev if std_dev is not None else 1.0
        # )
    elif method == "LogNormal":
        dist_arr = _gen_lognormal(
            dimension, num_points, std_dev if std_dev is not None else 1.0
        )
        # return _gen_lognormal(
        #     dimension, num_points, std_dev if std_dev is not None else 1.0
        # )
    else:
        raise ValueError(f"Unknown sampling method: {method}")
    # Shuffle the array to avoid the correlation between the parameters
    return np.random.permutation(dist_arr.flatten()).reshape(dist_arr.shape)


def sample_param_df(prange_df: pd.DataFrame, num_params: int = 2**10) -> pd.DataFrame:
    """
    Samples parameter values based on the provided parameter range DataFrame.

    This function takes a DataFrame containing parameter ranges and sampling
    methods, and generates a new DataFrame with sampled parameter values. The
    sampling is performed according to the specified methods and standard
    deviations (if provided).

    Parameters
    ----------
    prange_df : pd.DataFrame
        A DataFrame containing parameter ranges and sampling methods. It must
        include at least the columns "Parameter" and "Sampling". Optionally, it
        can include "StdDev", "Minimum", and "Maximum" columns.
    num_params : int, optional
        The number of parameter samples to generate for each parameter. Default
        is 1024 (2**10).

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the sampled parameter values. The columns are
        ordered according to the original order of parameters in `prange_df`.
    """
    # Save the original order of parameters
    original_order = prange_df["Parameter"].tolist()
    # Sort the DataFrame by "Sampling" and "StdDev" columns
    sort_cols = ["Sampling"] + (["StdDev"] if "StdDev" in prange_df.columns else [])
    prange_df = prange_df.sort_values(by=sort_cols)
    # Get the parameter names in the order they appear in the sampling method sorted DataFrame
    param_order = prange_df["Parameter"].tolist()
    # Dictionary to store sampled values for each parameter
    sampled_values = {}

    # Group by sampling settings (including StdDev if available)
    group_cols = ["Sampling"] + (["StdDev"] if "StdDev" in prange_df.columns else [])
    for _, group in prange_df.groupby(group_cols, sort=False):
        method = group["Sampling"].iloc[0]
        std_val = group["StdDev"].iloc[0] if "StdDev" in group.columns else None
        dims = group.shape[0]
        samples = sample_distribution(method, dims, num_params, std_dev=std_val)
        # Assign each sampled column to its corresponding parameter name.
        for i, param in enumerate(group["Parameter"]):
            sampled_values[param] = samples[:, i]
    # Create a DataFrame from the sampled values
    sampled_df = pd.DataFrame({p: sampled_values[p] for p in param_order})

    # Scale the sampled values to the specified range
    def scale_col(col: pd.Series, param_name: str) -> pd.Series:
        row = prange_df[prange_df["Parameter"] == param_name].iloc[0]
        # The InhFld parameter is scaled differently and then inverted
        if "InhFld" in param_name:
            # Change the min and max values to the reciprocal
            row_min, row_max = 1 / row["Maximum"], 1 / row["Minimum"]
            # Invert the scaled values and return the series
            return pd.Series(1 / scale_array(col.values, row_min, row_max))
        return pd.Series(
            scale_array(
                col.values,
                row["Minimum"],
                row["Maximum"],
                round_int=("Hill" in param_name),
            )
        )

    # Scale the sampled values for each parameter
    for col in sampled_df.columns:
        sampled_df[col] = scale_col(sampled_df[col], col)
    # Return the sampled DataFrame with the original parameter order
    return sampled_df[original_order]


# Functionto calculate the median g/k values based on the incoming edges. This function will be used by the get_thr_ranges function to get the proper threshold ranges which agrees with the half functional rule.
def _get_updated_gkn_hills(
    gk_n: np.ndarray,
    in_edge_params: pd.DataFrame,
    in_edge_topo: pd.DataFrame,
    num_params: int = 2**10,
) -> float:
    """
    Update the g/k values for a node using incoming edges and their Hills equations. Used by the get_thr_ranges function to generate threshold ranges which follow half functional rule.

    Parameters:
        gk_n (np.ndarray): Array of g/k values for the node.
        in_edge_params (pd.DataFrame): DataFrame containing parameters for incoming edges.
        in_edge_topo (pd.DataFrame): DataFrame containing the topology of incoming edges.
        num_params (int, optional): Number of parameter samples to generate. Default is 1024.

    Returns:
        float: The median of the product of updated g/k values across all incoming edges.
    """
    # print(f"Incoming edge parameters:\n{in_edge_params}")
    # Create a DataFrame to store the updated g/k values
    gk_hills = pd.DataFrame({"src_gk_n": gk_n})
    # Sample parameters for incoming edges
    inedg_param_samples = sample_param_df(in_edge_params, num_params)
    # Calculate the updated g/k values based on the Hills equation for each incoming edge
    for _, row in in_edge_topo.iterrows():
        # Get the source and target nodes, and the regulation type
        insrc_nd, tgt_nd, type_reg = row["Source"], row["Target"], row["Type"]
        # print(
        #     f"Processing incoming edge from {insrc_nd} to {tgt_nd} of type {type_reg}"
        # )
        # Calculate the updated g/k value based on the regulation type
        if type_reg == 1:
            fld = inedg_param_samples[f"ActFld_{insrc_nd}_{tgt_nd}"].values
            g = inedg_param_samples[f"Prod_{insrc_nd}"].values
            k = inedg_param_samples[f"Deg_{insrc_nd}"].values
            n = inedg_param_samples[f"Hill_{insrc_nd}_{tgt_nd}"].values
            thr = inedg_param_samples[f"Thr_{insrc_nd}_{tgt_nd}"].values
            gk_hills[f"AcH_{insrc_nd}_{tgt_nd}"] = psH(g / k, fld, thr, n)
        else:
            fld = inedg_param_samples[f"InhFld_{insrc_nd}_{tgt_nd}"].values
            g = inedg_param_samples[f"Prod_{insrc_nd}"].values
            k = inedg_param_samples[f"Deg_{insrc_nd}"].values
            n = inedg_param_samples[f"Hill_{insrc_nd}_{tgt_nd}"].values
            thr = inedg_param_samples[f"Thr_{insrc_nd}_{tgt_nd}"].values
            gk_hills[f"InH_{insrc_nd}_{tgt_nd}"] = nsH(g / k, fld, thr, n)
    # Calculate the median of the product of updated g/k values across all incoming edges and return the median g/k value
    return gk_hills.prod(axis=1).median()


# Function to get the threshold ranges for a given source node. This function will be used by the add_thr_rows function to add threshold-related rows to the parameter range DataFrame.
def get_thr_ranges(
    source_node: str,
    topo_df: pd.DataFrame,
    prange_df: pd.DataFrame,
    num_params: int = 2**10,
    # optimise: bool = False,
) -> float:
    """
    Calculate the median threshold range for a given source node.

    Parameters
    ----------
    source_node : str
        The source node for which the threshold range is calculated.
    topo_df : pd.DataFrame
        DataFrame containing the topology information.
    prange_df : pd.DataFrame
        DataFrame containing the parameter ranges.
    num_params : int, optional
        Number of parameters to sample. Defaults to 1024.

    Returns
    -------
    float
        The median threshold range for the given source node.
    """
    # Get the production and degradation rates for the source node
    sn_params = prange_df[
        prange_df["Parameter"].str.contains(f"Prod_{source_node}|Deg_{source_node}")
    ]
    # Sample the production and degradation rates for the source node
    sn_gk = sample_param_df(sn_params, num_params)
    # Calculate the g/k values for the source node
    sn_gk_n = (sn_gk[f"Prod_{source_node}"] / sn_gk[f"Deg_{source_node}"]).to_numpy()
    # Get the incoming edges for the source node
    in_edge_topo = topo_df[topo_df["Target"] == source_node]
    # If there are incoming edges, calculate the updated g/k values based on the Hills equation
    if not in_edge_topo.empty:
        isn = "|".join(in_edge_topo["Source"].unique())
        # Get the parameters for the incoming edges
        in_edge_params = prange_df[
            prange_df["Parameter"].str.contains(
                f"Fld_{isn}_{source_node}|Thr_{isn}_{source_node}|Hill_{isn}_{source_node}"
            )
            | prange_df["Parameter"].str.contains(f"Prod_{isn}|Deg_{isn}")
        ]
        # Sample the parameters for the incoming edges
        isn_gk = sample_param_df(
            in_edge_params[in_edge_params["Parameter"].str.contains("Prod_|Deg_")],
            num_params,
        )
        # Calculate the updated g/k values based on the Hills equation for the incoming edges
        for in_node in in_edge_topo["Source"].unique():
            # print(f"Processing incoming edge from {in_node} to {source_node}")
            in_gk = isn_gk[f"Prod_{in_node}"] / isn_gk[f"Deg_{in_node}"]
            in_gk_median = np.median(in_gk)
            # print(f"Median g/k value for {in_node}: {in_gk_median}")
            in_edge_params.loc[
                in_edge_params["Parameter"].str.contains(
                    f"Thr_{in_node}_{source_node}"
                ),
                ["Minimum", "Maximum"],
            ] = [0.02 * in_gk_median, 1.98 * in_gk_median]
        # Update the g/k values for the source node based on the incoming edges and return the median g/k value
        return _get_updated_gkn_hills(sn_gk_n, in_edge_params, in_edge_topo, num_params)
    else:
        # If there are no incoming edges, return the median g/k value for the source node
        return np.median(sn_gk[f"Prod_{source_node}"] / sn_gk[f"Deg_{source_node}"])


# Function to add threshold-related rows to the parameter range DataFrame. This function will be used by the get_param_range_df function to generate a parameter range DataFrame from the topology DataFrame.
def add_thr_rows(
    prange_df: pd.DataFrame, topo_df: pd.DataFrame, num_params: int = 2**10
) -> pd.DataFrame:
    """
    This function modifies the given parameter range DataFrame (`prange_df`) by adding
    threshold-related rows based on the topology DataFrame (`topo_df`). It calculates
    the median threshold values for source nodes and adjusts the minimum and maximum
    values for the corresponding parameters. Additionally, it scales the production
    parameters for the source nodes.

    Parameters
    ----------
    prange_df : pd.DataFrame
        The DataFrame containing parameter ranges.
    topo_df : pd.DataFrame
        The DataFrame containing the network topology.
    num_params : int, optional
        The number of parameters to consider. Default is 1024.

    Returns
    -------
    pd.DataFrame
        The modified parameter range DataFrame with added threshold-related rows.
    """
    # Get the parameter names, target nodes, and source nodes
    param_names, target_nodes, source_nodes = gen_param_names(topo_df)
    # Iterate over the unique source and target nodes
    for sn in set(target_nodes + source_nodes):
        # print(f"Processing node: {sn}")
        # If the node is a source node, its aplification factor is calculated if the median threshold value is less than 0.01
        if sn in source_nodes:
            median_thr_val = get_thr_ranges(sn, topo_df, prange_df, num_params)
            # print(f"Median threshold value for {sn}: {median_thr_val}")
            amplify_val = 1.0
            # Calculate the amplification factor if the median threshold value is less than 0.01
            if (median_thr_val * 0.02) < 0.010:
                exp_val = np.floor(np.log10(np.abs(median_thr_val * 0.02))).astype(int)
                amplify_val = 10 ** (-exp_val - 2)
            # Update the production and threshold values for the source node
            prange_df.loc[
                prange_df["Parameter"].str.contains(f"Thr_{sn}"), "Minimum"
            ] = median_thr_val * 0.02 * amplify_val
            prange_df.loc[
                prange_df["Parameter"].str.contains(f"Thr_{sn}"), "Maximum"
            ] = median_thr_val * 1.98 * amplify_val
            prange_df.loc[
                prange_df["Parameter"] == f"Prod_{sn}", ["Minimum", "Maximum"]
            ] *= amplify_val
    return prange_df


# Function to generate the parameter range dataframe from the topology of the network.
def gen_param_range_df(
    topo_df: pd.DataFrame,
    num_params: int = 2**10,
    sampling_method: Union[str, dict] = "Sobol",
    thr_rows: bool = True,
) -> pd.DataFrame:
    """
    Generate a parameter range DataFrame from the topology DataFrame.

    Parameters
    ----------
    topo_df : pd.DataFrame
        The topology DataFrame containing the network structure.
    num_params : int, optional
        The number of parameters to generate. Default is 1024.
    sampling_method : Union[str, dict], optional
        The sampling method to use. Can be a string specifying a single method for all parameters or a dictionary specifying methods for individual parameters. Default is "Sobol".
    thr_rows : bool, optional
        Whether to add threshold rows to the DataFrame. Default is True.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the parameter ranges and sampling methods.
    """
    # Generate the parameter names, target nodes, and source nodes
    param_names, target_nodes, source_nodes = gen_param_names(topo_df)
    # Create a DataFrame to store the parameter ranges
    prange_df = pd.DataFrame({"Parameter": param_names})
    # Set the default minimum and maximum values for the parameters
    prange_df.loc[
        prange_df["Parameter"].str.contains("Prod_"), ["Minimum", "Maximum"]
    ] = [1.0, 100.0]
    prange_df.loc[
        prange_df["Parameter"].str.contains("Deg_"), ["Minimum", "Maximum"]
    ] = [0.1, 1.0]
    prange_df.loc[
        prange_df["Parameter"].str.contains("ActFld_"), ["Minimum", "Maximum"]
    ] = [1.0, 100.0]
    prange_df.loc[
        prange_df["Parameter"].str.contains("InhFld_"), ["Minimum", "Maximum"]
    ] = [0.01, 1.0]
    prange_df.loc[
        prange_df["Parameter"].str.contains("Hill"), ["Minimum", "Maximum"]
    ] = [1.0, 6.0]
    # Set the sampling method for each parameter, if the sampling method is a dictionary set the specific method for the specified parameters
    if isinstance(sampling_method, str):
        prange_df["Sampling"] = sampling_method
        if sampling_method in ["Normal", "LogNormal"]:
            prange_df["StdDev"] = 1.0
    else:
        for param, method in sampling_method.items():
            prange_df.loc[prange_df["Parameter"].str.contains(param), "Sampling"] = (
                method
            )
        # If the sampling method is not specified for a parameter, set it to "Sobol"
        prange_df["Sampling"] = prange_df["Sampling"].fillna("Sobol")
        if any(prange_df["Sampling"].isin(["Normal", "LogNormal"])):
            prange_df["StdDev"] = 1.0
    # Fill the threshold rows of the parameter range DataFrame
    if thr_rows:
        prange_df = add_thr_rows(prange_df, topo_df, num_params)
    return prange_df


def gen_param_df(
    prange_df: pd.DataFrame = None,
    num_params: int = 2**10,
    topo_df: pd.DataFrame = None,
    sampling_method: Union[str, dict] = "Sobol",
    thr_rows: bool = True,
) -> pd.DataFrame:
    """
    Generate the final parameter DataFrame by sampling parameters.
    Parameters are grouped by their 'Sampling' (and 'StdDev' if present) to ensure
    that parameters in the same group follow the same distribution in the higher dimensions.
    The final DataFrame columns are arranged in the same order as in prange_df.
    The sampling methods can be: 'Sobol', 'LHS', 'Uniform', 'LogUniform', 'Normal', 'LogNormal'.

    Parameters
    ----------
    prange_df : pd.DataFrame, optional
        DataFrame with columns ["Parameter", "Minimum", "Maximum", "Sampling", ...].
    num_params : int, optional
        Number of samples to generate per parameter. Default is 1024.
    topo_df : pd.DataFrame, optional
        DataFrame containing the network topology information.
    sampling_method : Union[str, dict], optional
        The sampling method to use. Can be a string specifying a single method for all parameters or a dictionary specifying methods for individual parameters. Default is "Sobol". The methods can be: 'Sobol', 'LHS', 'Uniform', 'LogUniform', 'Normal', 'LogNormal'.
    thr_rows : bool, optional
        Whether to add threshold rows to the DataFrame. Default is True.

    Returns
    -------
    pd.DataFrame
        DataFrame of sampled and scaled parameters.
    """
    # If the parameter range dataframe is not given
    if prange_df is None:
        # Generating the parameter range dataframe from the topology of the network
        prange_df = gen_param_range_df(
            topo_df=topo_df,
            num_params=num_params,
            sampling_method=sampling_method,
            thr_rows=thr_rows,
        )
    # # Get the ordered parameter names
    # ordered_params = prange_df["Parameter"].tolist()
    # # Dictionary to store sampled values for each parameter
    # sampled_dict = {}

    # # Group by 'Sampling' and 'StdDev' columns
    # grouping_cols = ["Sampling"]
    # if "StdDev" in prange_df.columns:
    #     grouping_cols.append("StdDev")

    # # Iterate over the groups and sample the parameters
    # for _, group in prange_df.groupby(grouping_cols, sort=False):
    #     method = group["Sampling"].iloc[0]
    #     std_val = group["StdDev"].iloc[0] if "StdDev" in group.columns else None
    #     dims = group.shape[0]
    #     samples = sample_distribution(method, dims, num_paras, std_dev=std_val)
    #     group_sorted = group.sort_index().reset_index(drop=True)
    #     for i, row in group_sorted.iterrows():
    #         param_name = row["Parameter"]
    #         min_val = row["Minimum"]
    #         max_val = row["Maximum"]
    #         col_samples = samples[:, i]
    #         if "Hill" in param_name:
    #             scaled = np.ceil(max_val * col_samples)
    #         elif "InhFld" in param_name:
    #             inv_min = 1 / max_val
    #             inv_max = 1 / min_val
    #             scaled = 1 / (inv_min + (inv_max - inv_min) * col_samples)
    #         else:
    #             scaled = min_val + (max_val - min_val) * col_samples
    #         sampled_dict[param_name] = scaled
    # # Create a DataFrame from the sampled values
    # data = {param: sampled_dict[param] for param in ordered_params}
    # # Return the DataFrame with the original parameter order
    # return pd.DataFrame(data, columns=ordered_params)
    # Use the sample_param_df function to sample the parameters
    param_df = sample_param_df(prange_df, num_params)
    # Add the ParamNum column to the DataFrame
    # param_df["ParamNum"] = param_df.index + 1
    param_df = param_df.assign(ParamNum=param_df.index + 1)
    return param_df


def gen_init_cond(topo_df: pd.DataFrame, num_init_conds: int = 2**10) -> pd.DataFrame:
    """
    Generate initial conditions for each node based on the topology.

    Parameters
    ----------
    topo_df : pd.DataFrame
        DataFrame containing the topology information.
    num_init_conds : int, optional
        Number of initial conditions to generate. Default is 2**10.

    Returns
    -------
    pd.DataFrame
        DataFrame containing the generated initial conditions for each node.
    """
    _, target_nodes, source_nodes = gen_param_names(topo_df)
    unique_nodes = sorted(set(source_nodes + target_nodes))
    init_conds = _gen_sobol_seq(len(unique_nodes), num_init_conds)
    # Scale initial conditions between 1 and 100
    init_conds = 1 + init_conds * (100 - 1)
    initcond_df = pd.DataFrame(init_conds, columns=unique_nodes)
    # A new columns for the intial condition numbers
    # initcond_df["InitCondNum"] = initcond_df.index + 1
    initcond_df = initcond_df.assign(InitCondNum=initcond_df.index + 1)
    return initcond_df



import numpy as np
import pandas as pd
import os
import glob
import jax.numpy as jnp
from jax import vmap, lax, jit, debug
from scipy.stats import qmc
import time


# Function to convert the dataframe into a numpy matrix
def parse_topo_to_matrix(topofile_path):
    """
    Parses a topology file into an adjacency matrix.

    This function reads a topology file, and converts it into an adjacency matrix.
    The adjacency matrix is then converted to a JAX array.

    Parameters
    ----------
    topofile_path : str
        The path to the topology file. The file should be in a format readable by pandas ```read_csv``` with whitespace as the delimiter.

    Returns
    -------
    tuple
        A tuple containing:
        topo_adj : jax.numpy.ndarray
            The adjacency matrix as a JAX array.
        node_names : list
            A list of node names in the order they appear in the adjacency matrix.
    """
    # Read the topo file as a pandas dataframe
    topo_df = pd.read_csv(topofile_path, sep=r"\s+")
    # Get the node names
    node_names = set(
        list(topo_df["Source"].unique()) + list(topo_df["Target"].unique())
    )
    # Adding self loops with type 0 for all the nodes if self loop of that node is not present
    # This is done to ensure that the adjacency matrix is square i.e all the only target or source nodes are also included
    # Subsetting the dataframe to get the self loops
    self_loops = topo_df[topo_df["Source"] == topo_df["Target"]]
    # Getting the list of nodes without self loops
    nodes_without_self_loops = list(node_names - set(self_loops["Source"]))
    # Creating a dataframe for the nodes without self loops
    self_loops_df = pd.DataFrame(
        [
            nodes_without_self_loops,
            nodes_without_self_loops,
            [0] * len(nodes_without_self_loops),
        ]
    ).T
    # Renaming the columns
    self_loops_df.columns = ["Source", "Target", "Type"]
    # Adding the self loops to the topo dataframe
    topo_df = pd.concat([topo_df, self_loops_df], ignore_index=True)
    # Convert the type column to float
    topo_df["Type"] = topo_df["Type"].astype(float)
    # Replacing the 2s with -1s
    topo_df["Type"] = topo_df["Type"].replace({2: -1})
    # Pivot the dataframe to get the adjacencey matrix
    topo_df = topo_df.pivot(index="Source", columns="Target", values="Type")
    # Make the node names set into a list
    node_names = sorted(list(node_names))
    # Reorder the columns and rows
    topo_df = topo_df.loc[node_names, node_names]
    # Replace the NaN values with 0s
    topo_df.fillna(0, inplace=True)
    # Convert the dataframe to a numpy adjacency matrix
    topo_adj = topo_df.to_numpy()
    # Convert the adjacency matrix to a jax array
    topo_adj = jnp.array(topo_adj, dtype=jnp.int16)
    return topo_adj, node_names


# Function to sample and scale the distributions
def generate_intial_conditions(num_nodes, num_samples):
    """
    Generate initial conditions for a given number of nodes and samples.

    This function generates initial conditions using Sobol sequences and scales the generated samples to [0, 1].

    Parameters
    ----------
    num_nodes : int
        The number of nodes for which to generate initial conditions.
    num_samples : int
        The number of samples to generate.

    Returns
    -------
    jax.numpy.ndarray
        The generated initial conditions as a JAX array of integers.
    """

    # Internal function to generate sobol sequences
    def _gen_sobol_seq(dimensions, num_samples, optimise=False):
        """
        Generate Sobol sequence samples.

        Parameters
        ----------
        dimensions : int
            The number of dimensions for the Sobol sequence.
        num_samples : int
            The number of samples to generate.
        optimise : bool, optional
            Whether to use optimization for generation. Defaults to False.

        Returns
        -------
        numpy.ndarray
            The generated Sobol sequence samples.
        """
        if not optimise:
            samples = qmc.Sobol(d=dimensions, scramble=True).random(num_samples)
        else:
            # Optimisation leads to a significant slowdown in the generation
            # Use only if needed
            samples = qmc.Sobol(
                d=dimensions, scramble=True, optimization="lloyd"
            ).random(num_samples)
        return samples

    # Function which scales a given distribution to the required ranges
    def _scale_distribution(
        sample,
        minmax_vals,
        round_int=False,
    ):
        """
        Scale the given distribution to the required ranges.

        Parameters
        ----------
        sample : numpy.ndarray
            The distribution to be scaled.
        minmax_vals : tuple
            The minimum and maximum values for scaling.
        round_int : bool, optional
            Whether to round the values to the nearest integer. Defaults to False.

        Returns
        -------
        numpy.ndarray
            The scaled distribution.
        """
        min_vals = minmax_vals[:, 0]
        max_vals = minmax_vals[:, 1]
        if round_int:
            min_vals = min_vals - 1
        # Scaling the values for the log distributions
        sample = min_vals + (max_vals - min_vals) * (
            sample - np.min(sample, axis=0)
        ) / (np.max(sample, axis=0) - np.min(sample, axis=0))
        # If round is required, return the scaled and rounded values
        if round_int:
            # Rounding the values to the next integer
            sample = np.ceil(sample)
            # # If values are present below the minimum value, shift them to the minimum value
            # If values are present above the maximum value, shift them to the maximum value
            sample = np.clip(sample, min_vals + 1, max_vals)
            # # Convert the values to integers
            sample = sample.astype(int)
            return sample
        else:
            return sample

    # Generate the MinMax values
    minmax_vals = np.array([[0, 1]] * num_nodes)
    # Generate the Sobol sequences
    unscaled_samples = _gen_sobol_seq(len(minmax_vals), num_samples)
    # Scaling the samples
    scaled_samples = _scale_distribution(
        unscaled_samples, np.array(minmax_vals), round_int=True
    )
    # COnvert the jax array
    scaled_samples = jnp.array(scaled_samples, dtype=jnp.int16)
    return scaled_samples


@jit
def sync_eval_next_state(
    prev_state,
    topo_adj,
    replacement_values,  # Vector of three values
):
    """
    Evaluate the next state of a system synchronously based on the previous state,
    topology adjacency matrix, and replacement values.

    Parameters
    ----------
    prev_state : jnp.ndarray
        The previous state of the system.
    topo_adj : jnp.ndarray
        The topology adjacency matrix representing the connections between nodes in the system.
    replacement_values : jnp.ndarray
        A vector of two values used for replacement based on the computed state conditions. The values are: [value_if_negative, value_if_positive].
        Value of 0 is not included as it is assumed that the state will remain the same if the node evaluates to 0 in that step.
        prev_state (jnp.ndarray): The previous state of the system.

    Returns
    -------
    jnp.ndarray
        The new state of the system as an array of int16.
    """
    # Compute the state of the incoming links for the state
    new_state = jnp.dot(prev_state, topo_adj)
    # debug.print("Scaled state: {}", new_state)
    # Apply replacement values based on conditions
    new_state = jnp.where(
        new_state < 0,
        replacement_values[0],
        jnp.where(new_state > 0, replacement_values[1], prev_state),
    )
    # Convert the new state to int16
    new_state = new_state.astype(jnp.int16)
    return new_state


@jit
def simulate_sync_trajectory(
    initial_condition,
    topo_adj,
    replacement_values,
    max_steps,
):
    """
    Simulates a synchronous trajectory of a system based on the given initial condition, topology adjacency matrix, and replacement values.

    Parameters
    ----------
    initial_condition : jnp.ndarray
        The initial state of the system.
    topo_adj : jnp.ndarray
        The topology adjacency matrix representing the connections between nodes in the system.
    replacement_values : jnp.ndarray
        The values used to replace the states during the simulation. It is a vector of two values in the form [value_if_negative, value_if_positive].
    max_steps : jnp.arange
        The range of steps to simulate. The simulation will run for each step in the range.

    Returns
    -------
    jnp.ndarray
        A JAX array containing the states of the system at each step, with the initial condition included at the beginning. The array also includes a column for the step indices. All -1 values in the states are replaced with 0 if the replacement values are [-1, 1].
    """

    # Initialize states array
    def step_fn(carry, _):
        current_state = carry
        next_state = sync_eval_next_state(
            current_state,
            topo_adj,
            replacement_values,
        )
        return next_state, next_state

    # Run the simulation loop
    _, states = lax.scan(step_fn, initial_condition, xs=max_steps)
    # Add the initial condition to the states at the beginning
    states = jnp.concatenate((jnp.expand_dims(initial_condition, axis=0), states))
    # Add a column for the steps
    states = jnp.hstack((jnp.arange(states.shape[0]).reshape(-1, 1), states))
    # Make sure that states are a jax array
    states = jnp.array(states, dtype=jnp.int16)
    # For the results where the replacement values are -1 and 1, some nodes will have -1 values, but for all purposes those should be considered as 0 values, so turning all -1 values to 0
    states = jnp.where(states == -1, 0, states)
    return states


@jit
def async_eval_next_state(
    prev_state,
    topo_adj,
    replacement_values,  # Vector of two values
    update_index,  # Index of the node to update
):
    """
    Asynchronously evaluates the next state of a node in an Ising model.

    Parameters
    ----------
    prev_state : jnp.ndarray
        The previous state vector of the system.
    topo_adj : jnp.ndarray
        The adjacency matrix representing the topology of the system.
    replacement_values : jnp.ndarray
        A vector of two values used for state replacement based on conditions. The values of the vector should be [value_if_negative, value_if_positive]. The value for 0 is not included as it is assumed that the state will remain the same if the node evaluates to 0 in that step.
    update_index : int
        The index of the node to update.

    Returns
    -------
    jnp.ndarray
        The new state vector after updating the specified node.
    """
    # debug.print("Update index: {}", update_index)
    # debug.print("Prev state: {}", prev_state)
    # Compute the state update only for the selected index
    new_value = jnp.dot(prev_state, topo_adj[update_index])  # Only update one row
    # debug.print("New value0: {}", new_value)
    # Apply replacement values based on conditions
    new_value = jnp.where(
        new_value < 0,
        replacement_values[0],
        jnp.where(new_value > 0, replacement_values[1], prev_state[update_index]),
    ).astype(jnp.int16)
    # debug.print("New value1: {}", new_value)
    # Scatter the updated value into the state vector
    new_state = prev_state.at[update_index].set(new_value)
    # debug.print("New state: {}", new_state)
    return new_state


@jit
def simulate_async_trajectory(
    initial_condition,
    topo_adj,
    replacement_values,
    update_indices,  # Vector of indices specifying which node to update at each step
):
    """
    Simulates an asynchronous trajectory of a system given an initial condition and update indices.

    Parameters
    ----------
    initial_condition : jnp.ndarray
        The initial condition of the system.
    topo_adj : jnp.ndarray
        The adjacency matrix representing the topology of the system.
    replacement_values : jnp.ndarray
        The values used to replace the states during the simulation. It is a vector of two values in the form [value_if_negative, value_if_positive].
    update_indices : jnp.ndarray
        A vector of indices specifying which node to update at each step. The length of the vector should be equal to the number of steps. If not, the simulation will only run until the length of the update_indices.

    Returns
    -------
    jnp.ndarray
        A JAX array containing the states of the system at each step, with the initial condition included at the beginning. The array also includes a column for the step indices. All -1 values in the states are replaced with 0 if the replacement values are [-1, 1].
    """

    def step_fn(carry, update_index):
        next_state = async_eval_next_state(
            carry,
            topo_adj,
            replacement_values,
            update_index,
        )
        return next_state, next_state

    # Run the simulation loop, passing update_indices as xs
    _, states = lax.scan(step_fn, initial_condition, xs=update_indices)
    # Add the initial condition to the states at the beginning
    states = jnp.concatenate((jnp.expand_dims(initial_condition, axis=0), states))
    # Add a column for the steps
    states = jnp.hstack((jnp.arange(states.shape[0]).reshape(-1, 1), states))
    # Make sure that states are the right dtype
    states = jnp.array(states, dtype=jnp.int16)
    # For the results where the replacement values are -1 and 1, some nodes will have -1 values, but for all purposes those should be considered as 0 values, so turning all -1 values to 0
    states = jnp.where(states == -1, 0, states)
    return states


# Function to pack the 0/1 states into bits
@jit
def packbit_states(states):
    states = jnp.concatenate(
        [states[:, 0:1], jnp.packbits(states[:, 1:], axis=1)], axis=1
    )
    return states


def run_simulations(
    topo_file,
    num_initial_conditions=None,
    inital_conditions=None,
    max_steps=None,
    batch_size=None,
    replacement_values=jnp.array([0, 1]),
    mode="sync",
    packbits=False,
    save_dir="IsingSimulResults",
):
    """
    Run synchronous or asynchronous simulations for a given topology.

    Parameters
    ----------
    topo_file : str
        The path to the topology file.
    num_initial_conditions : int, optional
        The number of initial conditions to sample. If not provided, the default is 2**10.
    inital_conditions : jax.numpy.ndarray, optional
        The initial conditions matrix with the individual initial conditions as rows of the matrix. If provided, num_initial_conditions is ignored.
    max_steps : int, optional
        The maximum number of steps to simulate. If not provided, it is calculated to be 10 times the number of nodes.
    batch_size : int, optional
        The number of samples per batch. If not provided, the default is 2**10.
    replacement_values : jax.numpy.ndarray, optional
        The values used for replacement in the simulation. The default is [0, 1].
    mode : str, optional
        The simulation mode, either "sync" or "async". The default is "sync".
    packbits : bool, optional
        Whether to pack the 0/1 states into bits to reduce memory usage. The default is False.
    save_dir : str, optional
        The directory to save the simulation results. The default is "IsingSimulResults".

    Returns
    -------
    None
        The simulation results are saved to a parquet file in save_dir within a subdirectory named after the topology file.

    Example
    -------
    Run the synchronous simulation for a topology file:

        >>> run_simulations(
        ...     topo_file="TOPOS/ER_1000_0.1.topo",
        ...     num_initial_conditions=2**10,
        ...     max_steps=100,
        ...     batch_size=2**10,
        ...     replacement_values=jnp.array([0, 1]),
        ...     mode="sync",
        ...     packbits=True,
        ...     save_dir="IsingSimulResults",
        ... )

    This will save the simulation results to a parquet file in the directory "IsingSimulResults/ER_1000_0.1/ER_1000_0.1_sync_results.parquet".

    Similary, the asynchronous simulation can be run by setting mode="async".

    If the initial conditions matrix is provided, the num_initial_conditions parameter is ignored. In this case, the initial_conditions matrix should have the individual initial conditions as rows.
    If only specfic inital conditions are to be used, the initial conditions matrix can be provided with the individual initial conditions as rows of the matrix. This provides control over simulating specific pre-defined initial conditions.

        >>> initial_conditions = jnp.array([[0, 1, 0, 1], [1, 0, 1, 0], [0, 0, 1, 1]])
        >>> run_simulations(
        ...     topo_file="TOPOS/ER_1000_0.1.topo",
        ...     initial_conditions=initial_conditions,
        ...     max_steps=100,
        ...     batch_size=2**10,
        ...     replacement_values=jnp.array([0, 1]),
        ...     mode="sync",
        ...     packbits=True,
        ...     save_dir="IsingSimulResults",
        ... )

    For cases where the replacement values are not [0, 1], the replacement values should be provided as a jax array of length 2 with the first value less than the second.

        >>> replacement_values = jnp.array([-1, 1]) # Replacement values are -1 for negetive and 1 for positive
        >>> run_simulations(
        ...     topo_file="TOPOS/ER_1000_0.1.topo",
        ...     num_initial_conditions=2**10,
        ...     max_steps=100,
        ...     batch_size=2**10,
        ...     replacement_values=replacement_values,
        ...     mode="sync",
        ...     packbits=True,
        ...     save_dir="IsingSimulResults",
        ... )

    The results for [-1, 1] replacment values will also be converted to 0 for all the -1 or 0 values in the states and 1s will remain as 1s when saving to the file. This is important as otherwise the packbits  would not work.

    The packbits function used is jnp.packbits which packs the 0/1 states into bits to reduce memory usage. This is useful when the number of nodes is large and the number of steps is also large. The memory usase can be reduced by a factor of 8 by packing the states into bits. If packbits is not set to True, the states are saved as is.

        >>> run_simulations(
        ...     topo_file="TOPOS/ER_1000_0.1.topo",
        ...     num_initial_conditions=2**10,
        ...     max_steps=100,
        ...     batch_size=2**10,
        ...     replacement_values=jnp.array([0, 1]),
        ...     mode="sync",
        ...     packbits=False,
        ...     save_dir="IsingSimulResults",
        ... )

    The final dataframe which is written to the parquet file has the following columns for the packbits=False case:
    - Step: The step number for the simulation.
    - Node names: The names of the nodes in the network.
    If the packbits=True, the final dataframe has the following columns:
    - Step: The step number for the simulation.
    - Byte_i: The ith byte of the packed states for the simulation.
    The Byte_i columns are created based on the number of nodes in the network. For example, if there are 100 nodes, there will be 13 columns for the packed states. They can be unpacked using jnp.unpackbits to get the unpacked state values.
    The order of the nodes for the Byte_i columns is the same as the order of the nodes in the network after parsing the topology file using the parse_topo_to_matrix function. The order of the nodes is saved in a text file in the simulation directory for easy reference. The naming convention for the text file is toponame_node_names_order.csv.
    """
    # Create the save directory if it does not exist
    os.makedirs(save_dir, exist_ok=True)
    # Create a subdirectory for the topology file
    topo_name = topo_file.split("/")[-1].split(".")[0]
    sim_dir = f"{save_dir}/{topo_name}"
    os.makedirs(sim_dir, exist_ok=True)
    # Get the adjacency matrix and node names
    topo_adj, node_names = parse_topo_to_matrix(topo_file)
    print(f"Running {mode} simulations for the network: {topo_file}")
    # Default batch size if not specified
    if batch_size is None:
        batch_size = 2**10
    # Default max steps if not specified
    if max_steps is None:
        # Determine max steps based on number of nodes
        max_steps = 10 * len(node_names)
    # Check if Inital conditions matrix is provided
    if inital_conditions is None:
        # Default number of inital conditions if not specified
        if num_initial_conditions is None:
            num_initial_conditions = 2**10
        # Generate random initial conditions
        initial_conditions = generate_intial_conditions(
            len(node_names), num_initial_conditions
        )
    else:
        # Check if the initial conditions matrix has the correct number of nodes
        if inital_conditions.shape[1] != len(node_names):
            raise ValueError(
                f"Initial conditions matrix should have the same number of columns as the number of nodes in the network. Expected {len(node_names)} columns, but got {inital_conditions.shape[1]}."
            )
        num_initial_conditions = inital_conditions.shape[0]
    # Check if the replacement values provided have length 2 and that the first value is less than the second
    if len(replacement_values) != 2 or replacement_values[0] >= replacement_values[1]:
        raise ValueError(
            "Replacement values must be a jax array of length 2 with the first value less than the second."
        )
    # Initialize an empty numpy array to store the results
    if not packbits:
        results_array = np.empty((0, len(node_names) + 1), dtype=np.int16)
        # Create the column names for the dataframe
        df_cols = ["Step"] + node_names
    else:
        results_array = np.empty(
            (0, int(np.ceil(len(node_names) / 8)) + 1), dtype=np.int16
        )
        # Create the column names for the dataframe
        df_cols = ["Step"] + [
            f"Byte_{i}" for i in range(int(np.ceil(len(node_names) / 8)))
        ]
        # Create a text file with the node names seperated by a comma for easy reference
        with open(f"{sim_dir}/{topo_name}_node_names_order.csv", "w") as f:
            f.write(",".join(node_names))
    # Start the simulation timer
    start_time = time.time()
    # Run synchronous or asynchronous simulations
    if mode == "sync":
        # Run synchronous simulations in batches
        for batch in range(0, num_initial_conditions, batch_size):
            batch_initial_conditions = initial_conditions[batch : batch + batch_size]
            batch_results = vmap(
                lambda x: simulate_sync_trajectory(
                    x,
                    topo_adj,
                    replacement_values,
                    jnp.arange(max_steps),
                ),
                in_axes=0,
            )(batch_initial_conditions)
            # Stack the batch results into a single array
            batch_results = jnp.vstack(batch_results)
            # If packbits is True, pack the 0/1 states into bits to reduce memory usage
            if packbits:
                batch_results = packbit_states(batch_results)
            # Append the batch results to the main results array
            results_array = np.vstack((results_array, batch_results))

    elif mode == "async":
        # Generate random update indices for asynchronous updates
        update_indices_matrix = np.random.randint(
            0, len(node_names), (max_steps, num_initial_conditions)
        )
        update_indices_matrix = jnp.array(update_indices_matrix, dtype=jnp.int16)
        # Run asynchronous simulations in batches
        for batch in range(0, num_initial_conditions, batch_size):
            batch_initial_conditions = initial_conditions[batch : batch + batch_size]
            batch_update_indices = update_indices_matrix[:, batch : batch + batch_size]
            batch_states = vmap(
                lambda x, y: simulate_async_trajectory(
                    x,
                    topo_adj,
                    replacement_values,
                    y,
                ),
                in_axes=(0, 1),
            )(batch_initial_conditions, batch_update_indices)
            # Stack the batch results into a single array
            batch_states = jnp.vstack(batch_states)
            # If packbits is True, pack the 0/1 states into bits to reduce memory usage
            if packbits:
                batch_states = packbit_states(batch_states)
            # Append the batch results to the main results array
            results_array = np.vstack((results_array, batch_states))

    # Create a dataframe from the results array
    results_df = pd.DataFrame(results_array, columns=df_cols)
    # Saving the results to a parquet file
    results_df.to_parquet(
        f"{sim_dir}/{topo_name}_{mode}_ising_results.parquet", index=False
    )
    # End the simulation timer
    print(f"Simulation time for {mode} mode: {time.time() - start_time:.2f} seconds")
    return None


if __name__ == "__main__":
    pass
    # Specify the path to the topo file
    topo_folder = "TOPOS"

    # Get the list of all the topo files
    topo_files = sorted(glob.glob(f"{topo_folder}/*.topo"))
    print(topo_files)

    # Specify the replacement values
    replacement_values = jnp.array([0, 1])
    # replacement_values = jnp.array([0, 1])

    # Specify the number of steps to simulateys free for open source and community projects
    max_steps = 100
    print(f"Number of steps: {max_steps}")

    # Specify the number of initial conditions to simulate
    num_initial_conditions = 2**14
    print(f"Number of initial conditions: {num_initial_conditions}")

    # Specify the batch size for parallel evaluation
    batch_size = 2**10

    # Loop over all the topo files
    for topo_file in topo_files:
        # Get the adjacency matrix and node names
        topo_adj, node_names = parse_topo_to_matrix(topo_file)
        print(f"Topology: {topo_file}")
        # # Run the simulation
        run_simulations(
            topo_file=topo_file,
            num_initial_conditions=num_initial_conditions,
            # max_steps=max_steps,
            batch_size=batch_size,
            replacement_values=replacement_values,
            mode="sync",
            packbits=True,
        )
        run_simulations(
            topo_file=topo_file,
            num_initial_conditions=num_initial_conditions,
            # max_steps=max_steps,
            batch_size=batch_size,
            replacement_values=replacement_values,
            mode="async",
            packbits=True,
        )
        # break



from . import gen_diffrax_ode
from . import gen_params
from . import ising_bool
from . import logical_bool
from . import racipe_run
# from . import racipe
from . import reg_funcs

__all__ = [
    "gen_diffrax_ode",
    "gen_params",
    "ising_bool",
    "logical_bool",
    "racipe_run",
    # "racipe",
    "reg_funcs",
]
